from unittest.mock import Mock

from marge.gitlab import GET, Api
from marge.pipeline import Pipeline

INFO = {
    "id": 47,
    "status": "pending",
    "ref": "new-pipeline",
    "sha": "a91957a858320c0e17f3a0eca7cfacbff50ea29a",
    "web_url": "https://link/pipelines/87",
}


# pylint: disable=attribute-defined-outside-init
class TestPipeline:
    def setup_method(self, _method):
        self.api = Mock(Api)

    def test_pipelines_by_branch(self):
        api = self.api
        pl1, pl2 = INFO, dict(INFO, id=48)
        api.call = Mock(return_value=[pl1, pl2])

        result = Pipeline.pipelines_by_branch(
            project_id=1234, branch=INFO["ref"], api=api
        )
        api.call.assert_called_once_with(
            GET(
                "/projects/1234/pipelines",
                {"ref": INFO["ref"], "order_by": "id", "sort": "desc"},
            )
        )
        assert [pl.info for pl in result] == [pl1, pl2]

    def test_pipelines_by_merge_request(self):
        api = self.api
        pl1, pl2 = INFO, dict(INFO, id=48)
        api.call = Mock(return_value=[pl1, pl2])

        result = Pipeline.pipelines_by_merge_request(
            project_id=1234, merge_request_iid=1, api=api
        )
        api.call.assert_called_once_with(
            GET("/projects/1234/merge_requests/1/pipelines")
        )
        assert [pl.info for pl in result] == [pl2, pl1]

    def test_properties(self):
        pipeline = Pipeline(api=self.api, project_id=1234, info=INFO)
        assert pipeline.id == 47
        assert pipeline.project_id == 1234
        assert pipeline.status == "pending"
        assert pipeline.ref == "new-pipeline"
        assert pipeline.sha == "a91957a858320c0e17f3a0eca7cfacbff50ea29a"
        assert pipeline.web_url == "https://link/pipelines/87"



from typing import TYPE_CHECKING, Optional

from . import gitlab


class User(gitlab.Resource):
    @classmethod
    def myself(cls, api: gitlab.Api) -> "User":
        info = api.call(gitlab.GET("/user"))
        if TYPE_CHECKING:
            assert isinstance(info, dict)

        return cls(api, info)

    @property
    def id(self) -> int:
        result = self._info["id"]
        if TYPE_CHECKING:
            assert isinstance(result, int)
        return result

    @property
    def is_admin(self) -> bool:
        result = self.info.get("is_admin") is True
        if TYPE_CHECKING:
            assert isinstance(result, bool)
        return result

    @classmethod
    def fetch_by_id(cls, user_id: int, api: gitlab.Api) -> "User":
        info = api.call(gitlab.GET(f"/users/{user_id}"))
        if TYPE_CHECKING:
            assert isinstance(info, dict)
        return cls(api, info)

    @property
    def name(self) -> str:
        result = self.info["name"]
        if TYPE_CHECKING:
            assert isinstance(result, str)
        return result.strip()

    @property
    def username(self) -> str:
        result = self.info["username"]
        if TYPE_CHECKING:
            assert isinstance(result, str)
        return result

    @property
    def email(self) -> Optional[str]:
        """Only visible to admins and 'self'. Sigh."""
        result = self.info.get("email")
        if TYPE_CHECKING:
            if result is not None:
                assert isinstance(result, str)
        return result

    @property
    def state(self) -> str:
        result = self.info["state"]
        if TYPE_CHECKING:
            assert isinstance(result, str)
        return result



# pylint: disable=protected-access
import dataclasses
import textwrap
from datetime import timedelta
from unittest.mock import ANY, MagicMock, create_autospec, patch

import pytest

import marge.git
import marge.gitlab
import marge.interval
import marge.merge_request
import marge.project
import marge.user
from marge.hooks import run_hook
from marge.job import CannotMerge, Fusion, MergeJob, MergeJobOptions, SkipMerge


class TestJob:
    def _mock_merge_request(self, **options):
        return create_autospec(
            marge.merge_request.MergeRequest, spec_set=True, **options
        )

    def get_merge_job(self, **merge_kwargs):
        params = {
            "api": create_autospec(marge.gitlab.Api, spec_set=True),
            "user": create_autospec(marge.user.User, spec_set=True),
            "project": create_autospec(marge.project.Project, spec_set=True),
            "repo": create_autospec(marge.git.Repo, spec_set=True),
            "options": MergeJobOptions(),
        }
        params.update(merge_kwargs)
        return MergeJob(**params)

    def test_get_source_project_when_is_target_project(self):
        merge_job = self.get_merge_job()
        merge_request = self._mock_merge_request()
        merge_request.source_project_id = merge_job._project.id
        r_source_project = merge_job.get_source_project(merge_request)
        assert r_source_project is merge_job._project

    def test_get_source_project_when_is_fork(self):
        with patch("marge.job.Project") as project_class:
            merge_job = self.get_merge_job()
            merge_request = self._mock_merge_request()
            r_source_project = merge_job.get_source_project(merge_request)

            project_class.fetch_by_id.assert_called_once_with(
                merge_request.source_project_id, api=merge_job._api
            )
            assert r_source_project is not merge_job._project
            assert r_source_project is project_class.fetch_by_id.return_value

    def test_get_mr_ci_status(self):
        with patch("marge.job.Pipeline", autospec=True) as pipeline_class:
            pipeline_success = [MagicMock(sha="abc", status="success", id=12345)]
            pipeline_class.pipelines_by_branch.return_value = pipeline_success
            pipeline_class.pipelines_by_merge_request.return_value = pipeline_success
            merge_job = self.get_merge_job()
            merge_request = self._mock_merge_request(sha="abc")

            r_ci_status, _, r_pipeline_id = merge_job.get_mr_ci_status(merge_request)

            pipeline_class.pipelines_by_merge_request.assert_called_once_with(
                merge_request.target_project_id, merge_request.iid, merge_job._api
            )
            assert r_ci_status == "success"
            assert r_pipeline_id == 12345

    def test_ensure_mergeable_mr_not_assigned(self):
        merge_job = self.get_merge_job()
        merge_request = self._mock_merge_request(
            state="opened", draft=False, squash_on_merge=False
        )
        with pytest.raises(SkipMerge) as exc_info:
            merge_job.ensure_mergeable_mr(merge_request)
        assert (
            exc_info.value.reason
            == "The MR is not assigned to Marge anymore. Please assign it back to Marge to merge."
        )

    def test_dont_wait_for_ci_to_pass_if_running(self):
        with patch("marge.job.Pipeline", autospec=True) as pipeline_class:
            pipeline_running = [MagicMock(sha="abc", status="running", id=12345)]
            pipeline_class.pipelines_by_branch.return_value = pipeline_running
            pipeline_class.pipelines_by_merge_request.return_value = pipeline_running

            merge_job = self.get_merge_job(
                options=MergeJobOptions(skip_incomplete_pipelines=True)
            )
            merge_request = self._mock_merge_request(
                state="opened", draft=False, squash=False, sha="abc"
            )
            with pytest.raises(SkipMerge) as exc_info:
                merge_job.wait_for_ci_to_pass(merge_request)
            assert (
                exc_info.value.reason
                == "Skipping merge because pipeline is still running."
            )

    def test_ensure_mergeable_mr_state_not_ok(self):
        merge_job = self.get_merge_job()
        merge_request = self._mock_merge_request(
            assignee_ids=[merge_job._user.id],
            state="merged",
            draft=False,
            squash_on_merge=False,
        )
        with pytest.raises(CannotMerge) as exc_info:
            merge_job.ensure_mergeable_mr(merge_request)
        assert exc_info.value.reason == "The merge request is already merged!"

    def test_ensure_mergeable_mr_not_approved(self):
        merge_job = self.get_merge_job()
        merge_request = self._mock_merge_request(
            assignee_ids=[merge_job._user.id],
            state="opened",
            draft=False,
            squash_on_merge=False,
        )
        merge_request.fetch_approvals.return_value.sufficient = False
        with pytest.raises(CannotMerge) as exc_info:
            merge_job.ensure_mergeable_mr(merge_request)

        merge_request.fetch_approvals.assert_called_once()
        assert "Insufficient approvals" in str(exc_info.value)

    def test_ensure_mergeable_mr_draft(self):
        merge_job = self.get_merge_job()
        merge_request = self._mock_merge_request(
            assignee_ids=[merge_job._user.id], state="opened", draft=True
        )
        merge_request.fetch_approvals.return_value.sufficient = True
        with pytest.raises(CannotMerge) as exc_info:
            merge_job.ensure_mergeable_mr(merge_request)

        assert exc_info.value.reason == (
            "It is not possible to merge MRs marked as Draft. "
            "Please undraft it before assigning to Marge next time."
        )

    def test_ensure_mergeable_mr_unresolved_discussion(self):
        merge_job = self.get_merge_job()
        merge_request = self._mock_merge_request(
            assignee_ids=[merge_job._user.id],
            state="opened",
            draft=False,
            blocking_discussions_resolved=False,
        )
        merge_request.fetch_approvals.return_value.sufficient = True
        with pytest.raises(CannotMerge) as exc_info:
            merge_job.ensure_mergeable_mr(merge_request)

        assert exc_info.value.reason == (
            "Merge requests which have unresolved discussions cannot be merged. "
            "Please address all feedback and mark discussions as resolved."
        )

    def test_ensure_mergeable_mr_squash_wanted_and_trailers(self):
        merge_job = self.get_merge_job(
            project=create_autospec(marge.project.Project, spec_set=True),
            options=MergeJobOptions(add_reviewers=True),
        )
        merge_request = self._mock_merge_request(
            assignee_ids=[merge_job._user.id],
            state="opened",
            draft=False,
            squash_on_merge=True,
        )
        with pytest.raises(CannotMerge) as exc_info:
            merge_job.ensure_mergeable_mr(merge_request)

        assert exc_info.value.reason == (
            "Merging requests marked as auto-squash is not possible due to configuration. "
            "Please disable squashing or talk with the maintainers about the commit tagging config."
        )

    def test_unassign_from_mr(self):
        merge_job = self.get_merge_job()
        merge_request = self._mock_merge_request()

        # when we are not the author
        merge_job.unassign_from_mr(merge_request)
        merge_request.assign_to.assert_called_once_with(merge_request.author_id)

        # when we are the author
        merge_request.author_id = merge_job._user.id
        merge_job.unassign_from_mr(merge_request)
        merge_request.unassign.assert_called_once()

    def test_fuse_using_rebase(self):
        merge_job = self.get_merge_job(options=MergeJobOptions(fusion=Fusion.rebase))
        branch_a = "A"
        branch_b = "B"

        merge_job.fuse(branch_a, branch_b)

        merge_job._repo.rebase.assert_called_once_with(
            branch_a, branch_b, source_repo_url=ANY, local=ANY
        )

    def test_fuse_using_merge(self):
        merge_job = self.get_merge_job(options=MergeJobOptions(fusion=Fusion.merge))
        branch_a = "A"
        branch_b = "B"

        merge_job.fuse(branch_a, branch_b)

        merge_job._repo.merge.assert_called_once_with(
            branch_a, branch_b, source_repo_url=ANY, local=ANY
        )

    def test_pipeline_msg_hook_is_called(self, tmp_path):
        hooks_dir = tmp_path / "hooks"
        hooks_dir.mkdir(exist_ok=True)
        hook_file = hooks_dir / "pipeline_message.py"

        hook_content = textwrap.dedent("""
        def main(pipeline_id, project_id):
            return "Hook executed for pipeline %s in project %s" % (pipeline_id, project_id)
        """)
        hook_file.write_text(hook_content, encoding="utf-8")

        pipeline_id = 123
        project_id = "hooks_project"
        hook_ret = run_hook(hooks_dir, "pipeline_message", pipeline_id, project_id)
        assert hook_ret is not None

        expected_result = (
            f"Hook executed for pipeline {pipeline_id} in project {project_id}"
        )
        assert hook_ret == expected_result

    def test_pipeline_msg_hook_no_file(self, tmp_path):
        hooks_dir = tmp_path / "hooks"
        hooks_dir.mkdir(exist_ok=True)

        pipeline_id = 123
        project_id = "hooks_project"
        hook_ret = run_hook(hooks_dir, "no_such_hook", pipeline_id, project_id)
        assert hook_ret is None

    def test_raises_needs_rebase(self):
        merge_job = self.get_merge_job()
        merge_request = self._mock_merge_request(detailed_merge_status="need_rebase")

        with pytest.raises(marge.job.NeedsRebase) as exc_info:
            merge_job.wait_for_merge_status_to_resolve(merge_request)

        assert str(exc_info.value) == "MR needs to be rebased."


class TestMergeJobOptions:
    def test_default(self):
        assert MergeJobOptions() == MergeJobOptions(
            add_tested=False,
            add_part_of=False,
            add_reviewers=False,
            reapprove=False,
            approval_timeout=timedelta(seconds=0),
            embargo=marge.interval.IntervalUnion.empty(),
            ci_timeout=timedelta(minutes=15),
            fusion=Fusion.rebase,
            use_no_ff_batches=False,
            use_merge_commit_batches=False,
            skip_ci_batches=False,
            guarantee_final_pipeline=False,
            custom_allowed_approvers=None,
            custom_required_approvals=0,
        )

    def test_default_ci_time(self):
        three_min = timedelta(minutes=3)
        assert MergeJobOptions(ci_timeout=three_min) == dataclasses.replace(
            MergeJobOptions(), ci_timeout=three_min
        )



import re
import urllib.parse
from typing import TYPE_CHECKING

from . import gitlab

GET = gitlab.GET


class Commit(gitlab.Resource):
    @classmethod
    def fetch_by_id(cls, project_id: int, sha: str, api: gitlab.Api) -> "Commit":
        info = api.call(GET(f"/projects/{project_id}/repository/commits/{sha}"))
        if TYPE_CHECKING:
            assert isinstance(info, dict)
        return cls(api, info)

    @classmethod
    def last_on_branch(cls, project_id: int, branch: str, api: gitlab.Api) -> "Commit":
        info = api.call(
            GET(
                f"/projects/{project_id}/repository/branches/"
                f'{urllib.parse.quote(branch, safe="")}'
            )
        )
        if TYPE_CHECKING:
            assert isinstance(info, dict)
        commit_info = info["commit"]
        return cls(api, commit_info)

    @property
    def id(self) -> str:
        result = self._info["id"]
        if TYPE_CHECKING:
            assert isinstance(result, str)
        return result

    @property
    def short_id(self) -> str:
        result = self.info["short_id"]
        if TYPE_CHECKING:
            assert isinstance(result, str)
        return result

    @property
    def title(self) -> str:
        result = self.info["title"]
        if TYPE_CHECKING:
            assert isinstance(result, str)
        return result

    @property
    def author_name(self) -> str:
        result = self.info["author_name"]
        if TYPE_CHECKING:
            assert isinstance(result, str)
        return result

    @property
    def author_email(self) -> str:
        result = self.info["author_email"]
        if TYPE_CHECKING:
            assert isinstance(result, str)
        return result

    @property
    def status(self) -> str:
        result = self.info["status"]
        if TYPE_CHECKING:
            assert isinstance(result, str)
        return result

    @property
    def reviewers(self) -> list[str]:
        return re.findall(
            r"^Reviewed-by: ([^\n]+)$", self.info["message"], re.MULTILINE
        )

    @property
    def testers(self) -> list[str]:
        return re.findall(r"^Tested-by: ([^\n]+)$", self.info["message"], re.MULTILINE)



import argparse
import contextlib
import datetime
import os
import re
import shlex
import tempfile
from unittest import mock

import pytest

import marge.bot as bot_module
import tests.gitlab_api_mock as gitlab_mock
from marge import app, interval, job
from tests.test_user import INFO as user_info


@contextlib.contextmanager
def config_file():
    content = """
add-part-of: true
add-reviewers: true
add-tested: true
branch-regexp: foo.*bar
ci-timeout: 5min
embargo: Friday 1pm - Monday 7am
git-timeout: 150s
gitlab-url: "http://foo.com"
impersonate-approvers: true
project-regexp: foo.*bar
ssh-key: KEY
"""
    with tempfile.NamedTemporaryFile(
        mode="w", prefix="config-file-"
    ) as tmp_config_file:
        try:
            tmp_config_file.write(content)
            tmp_config_file.flush()
            yield tmp_config_file.name
        finally:
            tmp_config_file.close()


@contextlib.contextmanager
def env(**kwargs):
    original = os.environ.copy()

    os.environ.clear()
    for key, value in kwargs.items():
        os.environ[key] = value

    yield

    os.environ.clear()
    for key, value in original.items():
        os.environ[key] = value


@contextlib.contextmanager
def main(cmdline=""):
    def api_mock(gitlab_url, auth_token):
        assert gitlab_url == "http://foo.com"
        assert auth_token in ("NON-ADMIN-TOKEN", "ADMIN-TOKEN")
        api = gitlab_mock.Api(
            gitlab_url=gitlab_url, auth_token=auth_token, initial_state="initial"
        )
        user_info_for_token = dict(user_info, is_admin=auth_token == "ADMIN-TOKEN")
        api.add_user(user_info_for_token, is_current=True)
        return api

    class DoNothingBot(bot_module.Bot):
        instance = None

        def start(self):
            assert self.__class__.instance is None
            self.__class__.instance = self

        @property
        def config(self):
            return self._config

    with (
        mock.patch("marge.bot.Bot", new=DoNothingBot),
        mock.patch("marge.gitlab.Api", new=api_mock),
    ):
        app.main(args=shlex.split(cmdline))
        the_bot = DoNothingBot.instance
        assert the_bot is not None
        yield the_bot


def test_default_values():
    with env(
        MARGE_AUTH_TOKEN="NON-ADMIN-TOKEN",
        MARGE_SSH_KEY="KEY",
        MARGE_GITLAB_URL="http://foo.com",
    ):
        with main() as bot:
            assert bot.user.info == user_info
            assert bot.config.project_regexp == re.compile(".*")
            assert bot.config.git_timeout == datetime.timedelta(seconds=120)
            assert bot.config.merge_opts == job.MergeJobOptions()
            assert bot.config.merge_order == "created_at"


def test_embargo():
    with env(
        MARGE_AUTH_TOKEN="NON-ADMIN-TOKEN",
        MARGE_SSH_KEY="KEY",
        MARGE_GITLAB_URL="http://foo.com",
    ):
        with main('--embargo="Fri 1pm-Mon 7am"') as bot:
            assert bot.config.merge_opts == job.MergeJobOptions(
                embargo=interval.IntervalUnion.from_human("Fri 1pm-Mon 7am")
            )


def test_rebase_remotely():
    with env(
        MARGE_AUTH_TOKEN="NON-ADMIN-TOKEN",
        MARGE_SSH_KEY="KEY",
        MARGE_GITLAB_URL="http://foo.com",
    ):
        with main("--rebase-remotely") as bot:
            assert bot.config.merge_opts != job.MergeJobOptions()
            assert bot.config.merge_opts == job.MergeJobOptions(
                fusion=job.Fusion.gitlab_rebase
            )


def test_use_merge_strategy():
    with env(
        MARGE_AUTH_TOKEN="NON-ADMIN-TOKEN",
        MARGE_SSH_KEY="KEY",
        MARGE_GITLAB_URL="http://foo.com",
    ):
        with main("--use-merge-strategy") as bot:
            assert bot.config.merge_opts != job.MergeJobOptions()
            assert bot.config.merge_opts == job.MergeJobOptions(fusion=job.Fusion.merge)


def test_add_tested():
    with env(
        MARGE_AUTH_TOKEN="NON-ADMIN-TOKEN",
        MARGE_SSH_KEY="KEY",
        MARGE_GITLAB_URL="http://foo.com",
    ):
        with main("--add-tested") as bot:
            assert bot.config.merge_opts != job.MergeJobOptions()
            assert bot.config.merge_opts == job.MergeJobOptions(add_tested=True)


def test_use_merge_strategy_and_add_tested_are_mutualy_exclusive():
    with env(
        MARGE_AUTH_TOKEN="NON-ADMIN-TOKEN",
        MARGE_SSH_KEY="KEY",
        MARGE_GITLAB_URL="http://foo.com",
    ):
        with pytest.raises(app.MargeBotCliArgError):
            with main("--use-merge-strategy --add-tested"):
                pass


def test_add_part_of():
    with env(
        MARGE_AUTH_TOKEN="NON-ADMIN-TOKEN",
        MARGE_SSH_KEY="KEY",
        MARGE_GITLAB_URL="http://foo.com",
    ):
        with main("--add-part-of") as bot:
            assert bot.config.merge_opts != job.MergeJobOptions()
            assert bot.config.merge_opts == job.MergeJobOptions(add_part_of=True)


def test_add_reviewers():
    with env(
        MARGE_AUTH_TOKEN="NON-ADMIN-TOKEN",
        MARGE_SSH_KEY="KEY",
        MARGE_GITLAB_URL="http://foo.com",
    ):
        with pytest.raises(AssertionError):
            with main("--add-reviewers") as bot:
                pass

    with env(
        MARGE_AUTH_TOKEN="ADMIN-TOKEN",
        MARGE_SSH_KEY="KEY",
        MARGE_GITLAB_URL="http://foo.com",
    ):
        with main("--add-reviewers") as bot:
            assert bot.config.merge_opts != job.MergeJobOptions()
            assert bot.config.merge_opts == job.MergeJobOptions(add_reviewers=True)


def test_rebase_remotely_option_conflicts():
    for conflicting_flag in [
        "--use-merge-strategy",
        "--add-tested",
        "--add-part-of",
        "--add-reviewers",
    ]:
        with env(
            MARGE_AUTH_TOKEN="NON-ADMIN-TOKEN",
            MARGE_SSH_KEY="KEY",
            MARGE_GITLAB_URL="http://foo.com",
        ):
            with pytest.raises(app.MargeBotCliArgError):
                with main(f"--rebase-remotely {conflicting_flag}"):
                    pass


def test_impersonate_approvers():
    with env(
        MARGE_AUTH_TOKEN="NON-ADMIN-TOKEN",
        MARGE_SSH_KEY="KEY",
        MARGE_GITLAB_URL="http://foo.com",
    ):
        with pytest.raises(AssertionError):
            with main("--impersonate-approvers"):
                pass

    with env(
        MARGE_AUTH_TOKEN="ADMIN-TOKEN",
        MARGE_SSH_KEY="KEY",
        MARGE_GITLAB_URL="http://foo.com",
    ):
        with main("--impersonate-approvers") as bot:
            assert bot.config.merge_opts != job.MergeJobOptions()
            assert bot.config.merge_opts == job.MergeJobOptions(reapprove=True)


def test_approval_reset_timeout():
    with env(
        MARGE_AUTH_TOKEN="NON-ADMIN-TOKEN",
        MARGE_SSH_KEY="KEY",
        MARGE_GITLAB_URL="http://foo.com",
    ):
        with main("--approval-reset-timeout 1m") as bot:
            assert bot.config.merge_opts != job.MergeJobOptions()
            assert bot.config.merge_opts == job.MergeJobOptions(
                approval_timeout=datetime.timedelta(seconds=60)
            )


def test_project_regexp():
    with env(
        MARGE_AUTH_TOKEN="NON-ADMIN-TOKEN",
        MARGE_SSH_KEY="KEY",
        MARGE_GITLAB_URL="http://foo.com",
    ):
        with main("--project-regexp='foo.*bar'") as bot:
            assert bot.config.project_regexp == re.compile("foo.*bar")


def test_ci_timeout():
    with env(
        MARGE_AUTH_TOKEN="NON-ADMIN-TOKEN",
        MARGE_SSH_KEY="KEY",
        MARGE_GITLAB_URL="http://foo.com",
    ):
        with main("--ci-timeout 5m") as bot:
            assert bot.config.merge_opts != job.MergeJobOptions()
            assert bot.config.merge_opts == job.MergeJobOptions(
                ci_timeout=datetime.timedelta(seconds=5 * 60)
            )


def test_deprecated_max_ci_time_in_minutes():
    with env(
        MARGE_AUTH_TOKEN="NON-ADMIN-TOKEN",
        MARGE_SSH_KEY="KEY",
        MARGE_GITLAB_URL="http://foo.com",
    ):
        with main("--max-ci-time-in-minutes=5") as bot:
            assert bot.config.merge_opts != job.MergeJobOptions()
            assert bot.config.merge_opts == job.MergeJobOptions(
                ci_timeout=datetime.timedelta(seconds=5 * 60)
            )


def test_git_timeout():
    with env(
        MARGE_AUTH_TOKEN="NON-ADMIN-TOKEN",
        MARGE_SSH_KEY="KEY",
        MARGE_GITLAB_URL="http://foo.com",
    ):
        with main("--git-timeout '150 s'") as bot:
            assert bot.config.git_timeout == datetime.timedelta(seconds=150)


def test_branch_regexp():
    with env(
        MARGE_AUTH_TOKEN="NON-ADMIN-TOKEN",
        MARGE_SSH_KEY="KEY",
        MARGE_GITLAB_URL="http://foo.com",
    ):
        with main("--branch-regexp='foo.*bar'") as bot:
            assert bot.config.branch_regexp == re.compile("foo.*bar")


def test_source_branch_regexp():
    with env(
        MARGE_AUTH_TOKEN="NON-ADMIN-TOKEN",
        MARGE_SSH_KEY="KEY",
        MARGE_GITLAB_URL="http://foo.com",
    ):
        with main("--source-branch-regexp='foo.*bar'") as bot:
            assert bot.config.source_branch_regexp == re.compile("foo.*bar")


def test_git_reference_repo():
    with env(
        MARGE_AUTH_TOKEN="NON-ADMIN-TOKEN",
        MARGE_SSH_KEY="KEY",
        MARGE_GITLAB_URL="http://foo.com",
    ):
        with main("--git-reference-repo='/foo/reference_repo'") as bot:
            assert bot.config.git_reference_repo == "/foo/reference_repo"


def test_merge_order_updated():
    with env(
        MARGE_AUTH_TOKEN="NON-ADMIN-TOKEN",
        MARGE_SSH_KEY="KEY",
        MARGE_GITLAB_URL="http://foo.com",
    ):
        with main("--merge-order='updated_at'") as bot:
            assert bot.config.merge_order == "updated_at"


def test_merge_order_assigned():
    with env(
        MARGE_AUTH_TOKEN="NON-ADMIN-TOKEN",
        MARGE_SSH_KEY="KEY",
        MARGE_GITLAB_URL="http://foo.com",
    ):
        with main("--merge-order='assigned_at'") as bot:
            assert bot.config.merge_order == "assigned_at"


def test_skip_incomplete_pipelines():
    with env(
        MARGE_AUTH_TOKEN="NON-ADMIN-TOKEN",
        MARGE_SSH_KEY="KEY",
        MARGE_GITLAB_URL="http://foo.com",
    ):
        with main("--skip-incomplete-pipelines") as bot:
            assert bot.config.merge_opts.skip_incomplete_pipelines is True


def test_time_interval():
    _900s = datetime.timedelta(0, 900)
    assert [app.time_interval(x) for x in ["15min", "15m", ".25h", "900s"]] == [
        _900s
    ] * 4
    assert app.time_interval("120") == datetime.timedelta(seconds=120)


@pytest.mark.parametrize("time_interval", ["120x", "this is not a time"])
def test_invalid_time_interval(time_interval: str) -> None:
    with pytest.raises(
        argparse.ArgumentTypeError,
        match=rf"Invalid time interval .*: {time_interval!r}",
    ):
        assert app.time_interval(time_interval)


def test_disabled_auth_token_cli_arg():
    with env(MARGE_SSH_KEY="KEY", MARGE_GITLAB_URL="http://foo.com"):
        with pytest.raises(app.MargeBotCliArgError):
            with main("--auth-token=ADMIN-TOKEN"):
                pass


def test_disabled_ssh_key_cli_arg():
    with env(MARGE_AUTH_TOKEN="NON-ADMIN-TOKEN", MARGE_GITLAB_URL="http://foo.com"):
        with pytest.raises(app.MargeBotCliArgError):
            with main("--ssh-key=KEY"):
                pass


def test_config_file():
    with config_file() as config_file_name:
        with env(MARGE_AUTH_TOKEN="ADMIN-TOKEN"):
            with main(f"--config-file={config_file_name}") as bot:
                admin_user_info = {**user_info}
                admin_user_info["is_admin"] = True
                assert bot.user.info == admin_user_info
                assert bot.config.merge_opts != job.MergeJobOptions()
                assert bot.config.merge_opts == job.MergeJobOptions(
                    embargo=interval.IntervalUnion.from_human("Fri 1pm-Mon 7am"),
                    add_tested=True,
                    add_part_of=True,
                    add_reviewers=True,
                    reapprove=True,
                    ci_timeout=datetime.timedelta(seconds=5 * 60),
                )
                assert bot.config.project_regexp == re.compile("foo.*bar")
                assert bot.config.git_timeout == datetime.timedelta(seconds=150)
                assert bot.config.branch_regexp == re.compile("foo.*bar")


def test_config_overwrites():
    with config_file() as config_file_name:
        with env(MARGE_CI_TIMEOUT="20min", MARGE_AUTH_TOKEN="ADMIN-TOKEN"):
            with main(f"--git-timeout=100s --config-file={config_file_name}") as bot:
                admin_user_info = {**user_info}
                admin_user_info["is_admin"] = True
                assert bot.user.info == admin_user_info
                assert bot.config.merge_opts != job.MergeJobOptions()
                assert bot.config.merge_opts == job.MergeJobOptions(
                    embargo=interval.IntervalUnion.from_human("Fri 1pm-Mon 7am"),
                    add_tested=True,
                    add_part_of=True,
                    add_reviewers=True,
                    reapprove=True,
                    ci_timeout=datetime.timedelta(seconds=20 * 60),
                )
                assert bot.config.project_regexp == re.compile("foo.*bar")
                assert bot.config.git_timeout == datetime.timedelta(seconds=100)
                assert bot.config.branch_regexp == re.compile("foo.*bar")



from unittest.mock import Mock

from marge.branch import Branch
from marge.gitlab import DELETE, Api


# pylint: disable=attribute-defined-outside-init
class TestBranch:
    def setup_method(self) -> None:
        self.api = Mock(Api)

    def test_delete_by_name_should_escape_branch_name(self) -> None:
        Branch.delete_by_name(
            project_id=923, branch="branch/with/slashes", api=self.api
        )
        self.api.call.assert_called_once_with(
            DELETE("/projects/923/repository/branches/branch%2Fwith%2Fslashes")
        )

    def test_delete_by_name_without_slashes(self) -> None:
        Branch.delete_by_name(
            project_id=923, branch="branch-without_slashes", api=self.api
        )
        self.api.call.assert_called_once_with(
            DELETE("/projects/923/repository/branches/branch-without_slashes")
        )



"""
An auto-merger of merge requests for GitLab
"""

import argparse
import contextlib
import datetime
import logging
import re
import sys
import tempfile
from collections.abc import Iterator
from typing import Optional, cast

import configargparse  # type: ignore[import-untyped]

from . import bot, gitlab, interval
from . import user as user_module


class MargeBotCliArgError(Exception):
    pass


def time_interval(str_interval: str) -> datetime.timedelta:
    re_match = re.match(r"\A([\d.]+) ?(h|m(?:in)?|s)?\Z", str_interval)
    if re_match:
        quant, unit = re_match.groups()
        unit = unit if unit else "s"
        translate = {"h": "hours", "m": "minutes", "min": "minutes", "s": "seconds"}
        if unit in translate:
            return datetime.timedelta(**{translate[unit]: float(quant)})
    raise configargparse.ArgumentTypeError(
        f"Invalid time interval (e.g. 12[s|min|h]): {str_interval!r}"
    )


def _parse_config(args: list[str]) -> argparse.Namespace:
    # pylint: disable=too-many-statements
    def regexp(str_regex: str) -> "re.Pattern[str]":
        try:
            return re.compile(str_regex)
        except re.error as err:
            raise configargparse.ArgumentTypeError(
                f"Invalid regexp: {str_regex!r} ({err.msg})"
            )

    parser = configargparse.ArgParser(
        auto_env_var_prefix="MARGE_",
        ignore_unknown_config_file_keys=True,  # Don't parse unknown args
        config_file_parser_class=configargparse.YAMLConfigFileParser,
        formatter_class=configargparse.ArgumentDefaultsRawHelpFormatter,
        description=__doc__,
    )

    # NOTE: Use a "." and \n at the end of the help for each option.
    parser.add_argument(
        "--config-file",
        env_var="MARGE_CONFIG_FILE",
        type=str,
        is_config_file=True,
        help="Config file path.\n",
    )
    auth_token_group = parser.add_mutually_exclusive_group(required=True)
    auth_token_group.add_argument(
        "--auth-token",
        type=str,
        metavar="TOKEN",
        help=(
            "Your GitLab token.\n"
            "DISABLED because passing credentials on the command line is insecure:\n"
            'You can still set it via ENV variable or config file, or use "--auth-token-file" flag.\n'
        ),
    )
    auth_token_group.add_argument(
        "--auth-token-file",
        type=configargparse.FileType("rt"),
        metavar="FILE",
        help="Path to your GitLab token file.\n",
    )
    parser.add_argument(
        "--gitlab-url",
        type=str,
        required=True,
        metavar="URL",
        help='Your GitLab instance, e.g. "https://gitlab.example.com".\n',
    )
    repo_access = parser.add_mutually_exclusive_group(required=True)
    repo_access.add_argument(
        "--use-https",
        env_var="MARGE_USE_HTTPS",
        action="store_true",
        help="Use HTTP(S) instead of SSH for GIT repository access.\n",
    )
    repo_access.add_argument(
        "--ssh-key",
        type=str,
        metavar="KEY",
        help=(
            "The private ssh key for marge so it can clone/push.\n"
            "DISABLED because passing credentials on the command line is insecure:\n"
            'You can still set it via ENV variable or config file, or use "--ssh-key-file" flag.\n'
        ),
    )
    repo_access.add_argument(
        "--ssh-key-file",
        type=str,  # because we want a file location, not the content
        metavar="FILE",
        help="Path to the private ssh key for marge so it can clone/push.\n",
    )
    parser.add_argument(
        "--embargo",
        type=interval.IntervalUnion.from_human,
        metavar="INTERVAL[,..]",
        help='Time(s) during which no merging is to take place, e.g. "Friday 1pm - Monday 9am".\n',
    )
    experimental_group = parser.add_mutually_exclusive_group(required=False)
    experimental_group.add_argument(
        "--use-merge-strategy",
        action="store_true",
        help=(
            "Use git merge instead of git rebase to update the *source* branch (EXPERIMENTAL)\n"
            "If you need to use a strict no-rebase workflow (in most cases\n"
            "you don't want this, even if you configured gitlab to use merge requests\n"
            "to use merge commits on the *target* branch (the default).)\n"
        ),
    )
    parser.add_argument(
        "--rebase-remotely",
        action="store_true",
        help=(
            "Instead of rebasing in a local clone of the repository, use GitLab's\n"
            "built-in rebase functionality, via their API. Note that Marge can't add\n"
            "information in the commits in this case.\n"
        ),
    )
    parser.add_argument(
        "--add-tested",
        action="store_true",
        help='Add "Tested: marge-bot <$MR_URL>" for the final commit on branch after it passed CI.\n',
    )
    parser.add_argument(
        "--batch", action="store_true", help="Enable processing MRs in batches.\n"
    )
    parser.add_argument(
        "--add-part-of",
        action="store_true",
        help='Add "Part-of: <$MR_URL>" to each commit in MR.\n',
    )
    parser.add_argument(
        "--batch-branch-name",
        type=str,
        default="marge_bot_batch_merge_job",
        help="Branch name when batching is enabled.\n",
    )
    parser.add_argument(
        "--add-reviewers",
        action="store_true",
        help='Add "Reviewed-by: $approver" for each approver of MR to each commit in MR.\n',
    )
    parser.add_argument(
        "--keep-committers",
        action="store_true",
        help="Keep the original commit info during rebases.\n",
    )
    parser.add_argument(
        "--keep-reviewers",
        action="store_true",
        help='Ensure previous "Reviewed-by: $approver" aren\'t dropped by --add-reviewers\n',
    )
    parser.add_argument(
        "--impersonate-approvers",
        action="store_true",
        help="Marge-bot pushes effectively don't change approval status.\n",
    )
    parser.add_argument(
        "--merge-order",
        default="created_at",
        choices=("created_at", "updated_at", "assigned_at"),
        help="Order marge merges assigned requests. created_at (default), updated_at or assigned_at.\n",
    )
    parser.add_argument(
        "--approval-reset-timeout",
        type=time_interval,
        default="0s",
        help=(
            "How long to wait for approvals to reset after pushing.\n"
            'Only useful with the "new commits remove all approvals" option in a project\'s settings.\n'
            "This is to handle the potential race condition where approvals don't reset in GitLab\n"
            "after a force push due to slow processing of the event.\n"
        ),
    )
    parser.add_argument(
        "--project-regexp",
        type=regexp,
        default=".*",
        help="Only process projects that match; e.g. 'some_group/.*' or '(?!exclude/me)'.\n",
    )
    parser.add_argument(
        "--ci-timeout",
        type=time_interval,
        default="15min",
        help="How long to wait for CI to pass.\n",
    )
    parser.add_argument(
        "--max-ci-time-in-minutes",
        type=int,
        default=None,
        help="Deprecated; use --ci-timeout.\n",
    )
    parser.add_argument(
        "--git-timeout",
        type=time_interval,
        default="120s",
        help="How long a single git operation can take.\n",
    )
    parser.add_argument(
        "--git-reference-repo",
        type=str,
        default=None,
        help="A reference repo to be used when git cloning.\n",
    )
    parser.add_argument(
        "--branch-regexp",
        type=regexp,
        default=".*",
        help="Only process MRs whose target branches match the given regular expression.\n",
    )
    parser.add_argument(
        "--source-branch-regexp",
        type=regexp,
        default=".*",
        help="Only process MRs whose source branches match the given regular expression.\n",
    )
    parser.add_argument(
        "--debug",
        action="store_true",
        help="Debug logging (includes all HTTP requests etc).\n",
    )
    parser.add_argument(
        "--run-manual-jobs",
        action="store_true",
        help="Add this flag to have Marge run on manual jobs within the pipeline.\n",
    )
    parser.add_argument(
        "--use-no-ff-batches",
        action="store_true",
        help="Disable fast forwarding when merging MR batches.\n",
    )
    parser.add_argument(
        "--use-merge-commit-batches",
        action="store_true",
        help=(
            "Use merge commit when creating batches, so that the commits in the batch MR\n"
            "will be the same with in individual MRs. Requires sudo scope in the access token.\n"
        ),
    )
    parser.add_argument(
        "--skip-ci-batches",
        action="store_true",
        help="Skip CI when updating individual MRs when using batches.\n",
    )
    parser.add_argument(
        "--cli",
        action="store_true",
        help="Run marge-bot as a single CLI command, not a service.\n",
    )
    parser.add_argument(
        "--skip-incomplete-pipelines",
        action="store_true",
        help=(
            "Skips MRs with incomplete pipeline(s) instead of waiting for them\n"
            "and unassign when timeout occurs.\n"
            "This options is especially useful for CLI mode in scheduled pipeline.\n"
        ),
    )
    parser.add_argument(
        "--guarantee-final-pipeline",
        action="store_true",
        help="Guaranteed final pipeline when assigned to marge-bot.\n",
    )
    parser.add_argument(
        "--exc-comment",
        type=str,
        default=None,
        help="Provide additional text, like a log URL, to append to some exception-related MR comments.\n",
    )
    parser.add_argument(
        "--custom-approver",
        type=str,
        default=None,
        nargs="*",
        help=(
            "Specify one or more approver usernames to accept instead of asking GitLab.\n"
            "For CE approval use.\n"
        ),
    )
    parser.add_argument(
        "--custom-approvals-required",
        type=int,
        default=0,
        help=(
            "Required number of approvals from --custom-approval.\n"
            "For CE approval use.\n"
        ),
    )
    parser.add_argument(
        "--hooks-directory",
        type=str,
        default=None,
        help="Path to the directory where your custom hooks are located.\n",
    )
    parser.add_argument(
        "--sign-commits",
        action="store_true",
        help=(
            "If enabled, commits will be signed with the same SSH key used to authenticate."
        ),
    )

    config = cast(argparse.Namespace, parser.parse_args(args))

    if config.use_merge_strategy and config.batch:
        raise MargeBotCliArgError(
            "--use-merge-strategy and --batch are currently mutually exclusive"
        )
    if config.use_merge_strategy and config.add_tested:
        raise MargeBotCliArgError(
            "--use-merge-strategy and --add-tested are currently mutually exclusive"
        )
    if config.rebase_remotely:
        conflicting_flag = [
            "--use-merge-strategy",
            "--add-tested",
            "--add-reviewers",
            "--add-part-of",
            "--keep-committers",
            "--sign-commits",
        ]
        for flag in conflicting_flag:
            if getattr(config, flag[2:].replace("-", "_")):
                raise MargeBotCliArgError(
                    f"--rebase-remotely and {flag} are mutually exclusive"
                )

    cli_args = []
    # pylint: disable=protected-access
    for _, (_, value) in parser._source_to_settings.get(
        configargparse._COMMAND_LINE_SOURCE_KEY, {}
    ).items():
        cli_args.extend(value)
    for bad_arg in ["--auth-token", "--ssh-key"]:
        if any(bad_arg in arg for arg in cli_args):
            raise MargeBotCliArgError(
                f'"{bad_arg}" can only be set via ENV var or config file.'
            )
    return config


@contextlib.contextmanager
def _secret_auth_token_and_ssh_key(
    options: argparse.Namespace,
) -> Iterator[tuple[str, Optional[str]]]:
    auth_token = options.auth_token or options.auth_token_file.readline().strip()
    if options.use_https:
        yield auth_token, None
    elif options.ssh_key_file:
        yield auth_token, options.ssh_key_file
    else:
        with tempfile.NamedTemporaryFile(
            mode="w", prefix="ssh-key-"
        ) as tmp_ssh_key_file:
            try:
                tmp_ssh_key_file.write(options.ssh_key + "\n")
                tmp_ssh_key_file.flush()
                yield auth_token, tmp_ssh_key_file.name
            finally:
                tmp_ssh_key_file.close()


def main(args: Optional[list[str]] = None) -> int:
    if args is None:
        args = sys.argv[1:]
    logging.basicConfig()

    options = _parse_config(args)

    if options.debug:
        logging.getLogger().setLevel(logging.DEBUG)
    else:
        logging.getLogger("requests").setLevel(logging.WARNING)

    with _secret_auth_token_and_ssh_key(options) as (auth_token, ssh_key_file):
        if options.sign_commits and ssh_key_file is None:
            raise MargeBotCliArgError(
                "--sign-commits option is only possible when an SSH key is provided"
            )

        api = gitlab.Api(options.gitlab_url, auth_token)
        user = user_module.User.myself(api)
        if options.max_ci_time_in_minutes:
            logging.warning(
                "--max-ci-time-in-minutes is DEPRECATED, use --ci-timeout %dmin",
                options.max_ci_time_in_minutes,
            )
            options.ci_timeout = datetime.timedelta(
                minutes=options.max_ci_time_in_minutes
            )

        if options.batch:
            logging.warning("Experimental batch mode enabled")

        if options.use_merge_strategy:
            fusion = bot.Fusion.merge
        elif options.rebase_remotely:
            fusion = bot.Fusion.gitlab_rebase
        else:
            fusion = bot.Fusion.rebase

        config = bot.BotConfig(
            user=user,
            use_https=options.use_https,
            auth_token=auth_token,
            ssh_key_file=ssh_key_file,
            project_regexp=options.project_regexp,
            git_timeout=options.git_timeout,
            git_reference_repo=options.git_reference_repo,
            branch_regexp=options.branch_regexp,
            source_branch_regexp=options.source_branch_regexp,
            merge_order=options.merge_order,
            merge_opts=bot.MergeJobOptions(
                add_tested=options.add_tested,
                add_part_of=options.add_part_of,
                add_reviewers=options.add_reviewers,
                keep_committers=options.keep_committers,
                keep_reviewers=options.keep_reviewers,
                reapprove=options.impersonate_approvers,
                approval_timeout=options.approval_reset_timeout,
                embargo=options.embargo,
                ci_timeout=options.ci_timeout,
                fusion=fusion,
                use_no_ff_batches=options.use_no_ff_batches,
                use_merge_commit_batches=options.use_merge_commit_batches,
                skip_ci_batches=options.skip_ci_batches,
                guarantee_final_pipeline=options.guarantee_final_pipeline,
                custom_allowed_approvers=options.custom_approver,
                custom_required_approvals=options.custom_approvals_required,
                hooks_directory=options.hooks_directory,
                run_manual_jobs=options.run_manual_jobs,
                sign_commits=options.sign_commits,
                skip_incomplete_pipelines=options.skip_incomplete_pipelines,
            ),
            batch=options.batch,
            batch_branch_name=options.batch_branch_name,
            cli=options.cli,
            exc_comment=options.exc_comment,
        )

        marge_bot = bot.Bot(api=api, config=config)
        marge_bot.start()
    return 0



import datetime
import zoneinfo

import dateutil
import pytest
import time_machine

import marge.interval
from marge.interval import IntervalUnion, WeeklyInterval


def date(spec):
    return dateutil.parser.parse(spec)


class TestWeekly:
    def test_on_same_week(self):
        interval = WeeklyInterval(
            "Mon", datetime.time(10, 00), "Fri", datetime.time(18, 00)
        )
        assert interval.covers(date("Tuesday 3pm"))
        assert not interval.covers(date("Sunday 5pm"))

        assert interval.covers(date("Monday 10am"))
        assert not interval.covers(date("Monday 9:59am"))

        assert interval.covers(date("Friday 6pm"))
        assert not interval.covers(date("Friday 6:01pm"))

    def test_span_two_weeks(self):
        interval = WeeklyInterval(
            "Friday", datetime.time(12, 00), "Mon", datetime.time(7, 00)
        )
        assert interval.covers(date("Sunday 10am"))
        assert not interval.covers(date("Wed 10am"))

        assert interval.covers(date("Friday 12:00pm"))
        assert not interval.covers(date("Friday 11:59am"))

        assert interval.covers(date("Monday 7am"))
        assert not interval.covers(date("Monday 7:01am"))

    def test_from_human(self):
        working_hours = WeeklyInterval(
            "Mon", datetime.time(9, 00), "Fri", datetime.time(17, 0)
        )

        assert WeeklyInterval.from_human("Mon@9am - Fri@5pm") == working_hours
        assert WeeklyInterval.from_human("Monday 9:00 - Friday@17:00") == working_hours
        assert WeeklyInterval.from_human("Mon@9:00-Fri@17:00") == working_hours
        assert WeeklyInterval.from_human("Mon@9:00-Tue@17:00") != working_hours

    def test_from_human_with_timezone(self):
        working_hours = WeeklyInterval(
            "Mon", datetime.time(9, 00), "Fri", datetime.time(17, 0)
        )

        # During summer time
        now = datetime.datetime(2019, 8, 30, tzinfo=zoneinfo.ZoneInfo("Europe/London"))
        with time_machine.travel(now):
            assert (
                WeeklyInterval.from_human(
                    "Mon 10:00 Europe/London - Fri 18:00 Europe/London"
                )
                == working_hours
            )

        # Outside summer time
        now = datetime.datetime(2019, 12, 30, tzinfo=zoneinfo.ZoneInfo("Europe/London"))
        with time_machine.travel(now):
            assert (
                WeeklyInterval.from_human(
                    "Mon 09:00 Europe/London - Fri 17:00 Europe/London"
                )
                == working_hours
            )


class TestIntervalUnion:
    def test_empty(self):
        empty_interval = IntervalUnion.empty()
        assert empty_interval == IntervalUnion([])
        assert not empty_interval.covers(date("Monday 5pm"))

    def test_singleton(self):
        weekly = WeeklyInterval(
            "Mon", datetime.time(10, 00), "Fri", datetime.time(18, 00)
        )
        interval = IntervalUnion([weekly])
        assert interval.covers(date("Tuesday 3pm"))
        assert not interval.covers(date("Sunday 5pm"))

    def test_non_overlapping(self):
        weekly_1 = WeeklyInterval(
            "Mon", datetime.time(10, 00), "Fri", datetime.time(18, 00)
        )
        weekly_2 = WeeklyInterval(
            "Sat", datetime.time(12, 00), "Sun", datetime.time(9, 00)
        )
        interval = IntervalUnion([weekly_1, weekly_2])
        assert interval.covers(date("Tuesday 3pm"))
        assert not interval.covers(date("Saturday 9am"))
        assert interval.covers(date("Saturday 6pm"))
        assert not interval.covers(date("Sunday 11am"))

    def test_from_human(self):
        weekly_1 = WeeklyInterval(
            "Mon", datetime.time(10, 00), "Fri", datetime.time(18, 00)
        )
        weekly_2 = WeeklyInterval(
            "Sat", datetime.time(12, 00), "Sun", datetime.time(9, 00)
        )
        interval = IntervalUnion([weekly_1, weekly_2])

        assert interval == IntervalUnion.from_human(
            "Mon@10am - Fri@6pm,Sat@12pm-Sunday 9am"
        )
        assert IntervalUnion([weekly_1]) == IntervalUnion.from_human(
            "Mon@10am - Fri@6pm"
        )

    def test_from_human_with_timezone(self):
        weekly_1 = WeeklyInterval(
            "Mon", datetime.time(10, 00), "Fri", datetime.time(18, 00)
        )
        weekly_2 = WeeklyInterval(
            "Sat", datetime.time(12, 00), "Sun", datetime.time(9, 00)
        )
        interval = IntervalUnion([weekly_1, weekly_2])

        # During summer time
        now = datetime.datetime(2019, 8, 30, tzinfo=zoneinfo.ZoneInfo("Europe/London"))
        with time_machine.travel(now):
            assert (
                IntervalUnion.from_human(
                    "Mon 11:00 Europe/London - Fri 19:00 Europe/London,"
                    "Sat 13:00 Europe/London - Sun 10:00 Europe/London"
                )
                == interval
            )

        # Outside summer time
        now = datetime.datetime(2019, 12, 30, tzinfo=zoneinfo.ZoneInfo("Europe/London"))
        with time_machine.travel(now):
            assert (
                IntervalUnion.from_human(
                    "Mon 10:00 Europe/London - Fri 18:00 Europe/London,"
                    "Sat 12:00 Europe/London - Sun 09:00 Europe/London"
                )
                == interval
            )


class TestParseTime:
    invalid_time_msg = "Could not parse time string"

    def test_empty_string(self):
        with pytest.raises(ValueError, match=self.invalid_time_msg):
            marge.interval.parse_time("", "UTC")

    def test_no_colon(self):
        with pytest.raises(ValueError, match=self.invalid_time_msg):
            marge.interval.parse_time("0000", "UTC")

    def test_fewer_minutes(self):
        with pytest.raises(ValueError, match=self.invalid_time_msg):
            marge.interval.parse_time("00:0", "UTC")

    def test_extra_minutes(self):
        with pytest.raises(ValueError, match=self.invalid_time_msg):
            marge.interval.parse_time("00:000", "UTC")

    def test_surplus_minutes(self):
        with pytest.raises(ValueError, match="minute must be in 0..59"):
            marge.interval.parse_time("00:70", "UTC")

    def test_extra_hours(self):
        with pytest.raises(ValueError, match=self.invalid_time_msg):
            marge.interval.parse_time("000:00", "UTC")

    def test_surplus_hours(self):
        with pytest.raises(ValueError, match="hour must be in 0..23"):
            marge.interval.parse_time("30:00", "UTC")

    def test_12am_converts_to_zero(self):
        time = marge.interval.parse_time("12am", "UTC")
        assert time.hour == 0
        assert time.minute == 0

    def test_mixed_case_am(self):
        time = marge.interval.parse_time("12Am", "UTC")
        assert time.hour == 0
        assert time.minute == 0

    def test_mixed_case_pm(self):
        time = marge.interval.parse_time("12pM", "UTC")
        assert time.hour == 12
        assert time.minute == 0



# pylint: disable=too-many-locals,too-many-branches,too-many-statements
import dataclasses
import datetime
import enum
import logging as log
import time
from typing import TYPE_CHECKING, Any, Callable, Optional, cast

import requests

from marge.hooks import run_hook

from . import git, gitlab, interval
from .approvals import Approvals, CustomApprovals
from .branch import Branch
from .interval import IntervalUnion
from .merge_request import MergeRequest, MergeRequestRebaseFailed
from .pipeline import Pipeline
from .project import Project
from .user import User

GET, POST = gitlab.GET, gitlab.POST


class MergeJob:
    def __init__(
        self,
        *,
        api: gitlab.Api,
        user: User,
        project: Project,
        repo: git.Repo,
        options: "MergeJobOptions",
    ):
        self._api = api
        self._user = user
        self._project = project
        self._repo = repo
        self._options = options
        self._merge_timeout = datetime.timedelta(minutes=5)

        # Normally we trust GitLab to tell us if approval criteria are met
        self.approvals_factory: Callable[[gitlab.Api, dict[str, Any]], Approvals] = (
            Approvals
        )
        custom_allowed = options.custom_allowed_approvers
        if (
            custom_allowed is not None
            and custom_allowed
            and options.custom_required_approvals
        ):
            # But sometimes we provide this logic instead
            self.approvals_factory = lambda api, info: CustomApprovals(
                api,
                info,
                allowed_approvers=custom_allowed,
                approvals_required=options.custom_required_approvals,
            )

    @property
    def repo(self) -> git.Repo:
        return self._repo

    @property
    def project(self) -> Project:
        return self._project

    @property
    def opts(self) -> "MergeJobOptions":
        return self._options

    def execute(self, exc_comment: Optional[str] = None) -> None:
        raise NotImplementedError

    def ensure_mergeable_mr(self, merge_request: MergeRequest) -> None:
        merge_request.refetch_info()
        log.info("Ensuring MR !%s is mergeable", merge_request.iid)
        log.debug("Ensuring MR %r is mergeable", merge_request)

        if merge_request.draft:
            raise CannotMerge(
                "It is not possible to merge MRs marked as Draft."
                " Please undraft it before assigning to Marge next time."
            )

        if merge_request.squash_on_merge and self._options.requests_commit_tagging:
            raise CannotMerge(
                "Merging requests marked as auto-squash is not possible due to configuration."
                " Please disable squashing or talk with the maintainers about the commit tagging config."
            )

        approvals = merge_request.fetch_approvals(self.approvals_factory)
        if not approvals.sufficient:
            raise CannotMerge(
                "Insufficient approvals "
                f"(have: {approvals.approver_usernames} missing: {approvals.approvals_left})"
            )

        if not merge_request.blocking_discussions_resolved:
            raise CannotMerge(
                "Merge requests which have unresolved discussions cannot be merged."
                " Please address all feedback and mark discussions as resolved."
            )

        state = merge_request.state
        if state not in ("opened", "reopened", "locked"):
            if state in ("merged", "closed"):
                raise SkipMerge(f"The merge request is already {state}!")
            raise CannotMerge(f"The merge request is in an unknown state: {state}")

        if self.during_merge_embargo():
            raise SkipMerge(
                "Merging is temporarily disabled (embargo). Please try later."
            )

        if self._user.id not in merge_request.assignee_ids:
            raise SkipMerge(
                "The MR is not assigned to Marge anymore."
                " Please assign it back to Marge to merge."
            )

    def modify_commits(self, merge_request: MergeRequest) -> Optional[str]:
        log.info("Adding trailers for MR !%s", merge_request.iid)

        # add Reviewed-by
        should_add_reviewers = (
            self._options.add_reviewers
            and self._options.fusion is not Fusion.gitlab_rebase
        )
        reviewers = (
            _get_reviewer_names_and_emails(
                merge_request.fetch_commits(),
                merge_request.fetch_approvals(self.approvals_factory),
                self._api,
            )
            if should_add_reviewers
            else None
        )
        sha = None
        if reviewers is not None:
            sha = self._repo.tag_with_trailer(
                trailer_name="Reviewed-by",
                trailer_values=reviewers,
                branch=merge_request.source_branch,
                start_commit="origin/" + merge_request.target_branch,
                keep_trailers=self._options.keep_reviewers,
            )

        # add Tested-by
        should_add_tested = (
            self._options.add_tested
            and self._project.only_allow_merge_if_pipeline_succeeds
            and self._options.fusion is Fusion.rebase
        )

        tested_by = (
            [f"{self._user.name} <{merge_request.web_url}>"]
            if should_add_tested
            else None
        )
        if tested_by is not None:
            sha = self._repo.tag_with_trailer(
                trailer_name="Tested-by",
                trailer_values=tested_by,
                branch=merge_request.source_branch,
                start_commit=merge_request.source_branch + "^",
                keep_trailers=False,
            )

        # add Part-of
        should_add_parts_of = (
            self._options.add_part_of
            and self._options.fusion is not Fusion.gitlab_rebase
        )
        part_of = f"<{merge_request.web_url}>" if should_add_parts_of else None
        if part_of is not None:
            sha = self._repo.tag_with_trailer(
                trailer_name="Part-of",
                trailer_values=[part_of],
                branch=merge_request.source_branch,
                start_commit="origin/" + merge_request.target_branch,
                keep_trailers=False,
            )

        # sign commits
        if (
            self._options.sign_commits
            and self._options.fusion is not Fusion.gitlab_rebase
        ):
            sha = self._repo.sign_commits(
                branch=merge_request.source_branch,
                start_commit="origin/" + merge_request.target_branch,
            )
        return sha

    def get_mr_ci_status(
        self, merge_request: MergeRequest, commit_sha: Optional[str] = None
    ) -> tuple[Optional[str], Optional[str], Optional[int]]:
        if commit_sha is None:
            commit_sha = merge_request.sha

        pipelines = Pipeline.pipelines_by_merge_request(
            merge_request.target_project_id, merge_request.iid, self._api
        )
        current_pipeline = next(
            iter(pipeline for pipeline in pipelines if pipeline.sha == commit_sha), None
        )

        if current_pipeline:
            ci_status = current_pipeline.status
            pipeline_msg = f"See pipeline {current_pipeline.web_url}."
            pipeline_id = current_pipeline.id
        else:
            log.warning(
                "No pipeline listed for %s on branch %s",
                commit_sha,
                merge_request.source_branch,
            )
            ci_status = None
            pipeline_msg = "No pipeline associated."
            pipeline_id = None

        return ci_status, pipeline_msg, pipeline_id

    def wait_for_ci_to_pass(
        self, merge_request: MergeRequest, commit_sha: Optional[str] = None
    ) -> None:
        time_0 = datetime.datetime.utcnow()
        waiting_time_in_secs = 10

        if commit_sha is None:
            commit_sha = merge_request.sha

        log.info(
            "Waiting for CI to pass for MR !%s - %s",
            merge_request.iid,
            merge_request.web_url,
        )
        if TYPE_CHECKING:
            assert self._options.ci_timeout is not None
        consecutive_errors = 0
        while True:
            try:
                ci_status, pipeline_msg, pipeline_id = self.get_mr_ci_status(
                    merge_request, commit_sha=commit_sha
                )
            except (gitlab.InternalServerError, requests.exceptions.Timeout):
                consecutive_errors += 1
                if consecutive_errors > 5:
                    raise

                time.sleep(waiting_time_in_secs)
                continue

            if (
                self._options.hooks_directory
                and ci_status
                and pipeline_msg
                and pipeline_id
            ):
                hook_ret = run_hook(
                    self._options.hooks_directory,
                    "pipeline_message",
                    pipeline_id,
                    str(self._project.id),
                )
                if hook_ret is not None:
                    pipeline_msg += hook_ret

            if ci_status == "success":
                log.info("CI for MR !%s passed. %s", merge_request.iid, pipeline_msg)
                return

            if ci_status == "skipped":
                log.info("CI for MR !%s skipped. %s", merge_request.iid, pipeline_msg)
                return

            if ci_status == "failed":
                raise CannotMerge(f"CI failed! {pipeline_msg}")

            if ci_status in ("canceling", "canceled"):
                raise CannotMerge(f"The CI run was canceled. {pipeline_msg}")

            if ci_status == "manual":
                if not self._options.run_manual_jobs:
                    raise CannotMerge(
                        "Manual Step encountered with run-manual-jobs set to False"
                    )
                if not pipeline_id:
                    raise CannotMerge("Cannot run manual jobs: Invalid Pipeline ID")
                self.play_pending_manual_jobs(merge_request, pipeline_id)

            if ci_status not in ("created", "pending", "running", "manual"):
                log.warning("Suspicious CI status: %r. %s", ci_status, pipeline_msg)

            if self._options.skip_incomplete_pipelines:
                log.info("Skipping merge because pipeline is still running.")

                raise SkipMerge("Skipping merge because pipeline is still running.")

            # Do we have enough time left to wait for another loop?
            time_after_wait = datetime.datetime.utcnow() + datetime.timedelta(
                seconds=waiting_time_in_secs
            )
            if time_after_wait - time_0 < self._options.ci_timeout:
                log.debug(
                    "Waiting for %s secs before polling CI status again",
                    waiting_time_in_secs,
                )
                time.sleep(waiting_time_in_secs)
                continue

            break

        raise CannotMerge(f"CI is taking too long. {pipeline_msg}")

    def wait_for_merge_status_to_resolve(self, merge_request: MergeRequest) -> None:
        """
        This function is for polling the async `detailed_merge_status` field in
        merge_request API response. We suspected that the lag `detailed_merge_status`
        prevents MRs to be merged, and the fix did work for some users.

        But we are not sure if this is the root cause and if this is a proper fix. As there're some
        evidence that suggest gitlab will always check the mergeability synchronously while merging MRs.
        See more https://github.com/smarkets/marge-bot/pull/265#issuecomment-724147901
        """
        attempts = 3
        waiting_time_in_secs = 5

        log.info(
            "Waiting for MR !%s to have detailed_merge_status mergeable",
            merge_request.iid,
        )
        for attempt in range(attempts):
            merge_request.refetch_info()
            detailed_merge_status = merge_request.detailed_merge_status

            if detailed_merge_status == "mergeable":
                log.info(
                    "MR !%s can be merged on attempt %d", merge_request.iid, attempt
                )
                return

            if detailed_merge_status in ("unchecked", "checking", "approvals_syncing"):
                log.info(
                    "MR !%s merge status is '%s' on attempt %d.",
                    merge_request.iid,
                    detailed_merge_status,
                    attempt,
                )
            elif detailed_merge_status == "need_rebase":
                log.info("MR !%s needs to be rebased.", merge_request.iid)
                merge_request.comment(
                    "A merge request may have skipped the queue, trying again..."
                )
                raise NeedsRebase("MR needs to be rebased.")
            else:
                log.info(
                    "MR !%s cannot be merged on attempt %d with status %s",
                    merge_request.iid,
                    attempt,
                    detailed_merge_status,
                )
                raise CannotMerge(
                    f"MR has status '{detailed_merge_status}' and cannot be merged."
                )

            time.sleep(waiting_time_in_secs)

    def play_pending_manual_jobs(
        self, merge_request: MergeRequest, pipeline_id: int
    ) -> None:
        manual_job_list = Pipeline.manual_jobs_by_pipeline(
            merge_request.target_project_id, pipeline_id, self._api
        )
        for job in manual_job_list:
            if job["status"] == "manual" and job["allow_failure"] is False:
                self._api.call(
                    POST(
                        f"/projects/{merge_request.target_project_id}/jobs/{job['id']}/play"
                    )
                )

    def unassign_from_mr(self, merge_request: MergeRequest) -> None:
        log.info("Unassigning from MR !%s", merge_request.iid)
        author_id = merge_request.author_id
        if author_id is not None and author_id != self._user.id:
            merge_request.assign_to(author_id)
        else:
            merge_request.unassign()

    def during_merge_embargo(self) -> bool:
        now = datetime.datetime.utcnow()
        if TYPE_CHECKING:
            assert self.opts.embargo is not None
        return self.opts.embargo.covers(now)

    def maybe_reapprove(
        self, merge_request: MergeRequest, approvals: Approvals
    ) -> None:
        # Re-approve the merge request, in case us pushing it has removed approvals.
        if self.opts.reapprove:
            # approving is not idempotent, so we need to check first that there
            # are no approvals, otherwise we'll get a failure on trying to
            # re-instate the previous approvals
            def sufficient_approvals() -> bool:
                return merge_request.fetch_approvals(self.approvals_factory).sufficient

            # Make sure we don't race by ensuring approvals have reset since the push
            waiting_time_in_secs = 5
            if TYPE_CHECKING:
                assert self._options.approval_timeout is not None
            approval_timeout_in_secs = self._options.approval_timeout.total_seconds()
            iterations = round(approval_timeout_in_secs / waiting_time_in_secs)
            log.info("Checking if approvals have reset")
            while sufficient_approvals() and iterations:
                log.debug(
                    "Approvals haven't reset yet, sleeping for %s secs",
                    waiting_time_in_secs,
                )
                time.sleep(waiting_time_in_secs)
                iterations -= 1
            if not sufficient_approvals():
                approvals.reapprove()

    def fetch_source_project(
        self, merge_request: MergeRequest
    ) -> tuple[Project, Optional[str], str]:
        remote = "origin"
        remote_url = None
        source_project = self.get_source_project(merge_request)
        if source_project is not self._project:
            remote = "source"
            remote_url = source_project.ssh_url_to_repo
            self._repo.fetch(remote_name=remote, remote_url=remote_url)
        return source_project, remote_url, remote

    def get_source_project(self, merge_request: MergeRequest) -> Project:
        source_project = self._project
        if merge_request.source_project_id != self._project.id:
            source_project = Project.fetch_by_id(
                merge_request.source_project_id, api=self._api
            )
        return source_project

    def get_target_project(self, merge_request: MergeRequest) -> Project:
        return Project.fetch_by_id(merge_request.target_project_id, api=self._api)

    def fuse(
        self,
        source: str,
        target: str,
        source_repo_url: Optional[str] = None,
        local: bool = False,
    ) -> str:
        # NOTE: this leaves git switched to branch_a
        strategies = {
            Fusion.rebase: self._repo.rebase,
            Fusion.merge: self._repo.merge,
            Fusion.gitlab_rebase: self._repo.rebase,  # we rebase locally to know sha
        }

        strategy = strategies[self._options.fusion]
        return cast(
            str,
            strategy(
                source, target, source_repo_url=source_repo_url, local=local
            ),  # type: ignore[operator]
        )

    def update_from_target_branch_and_push(
        self,
        merge_request: MergeRequest,
        source_repo_url: Optional[str] = None,
        skip_ci: bool = False,
    ) -> tuple[str, str, str]:
        """Updates `source_branch` on `target_branch`, optionally add trailers and push.
        The update strategy can either be rebase or merge. The default is rebase.

        Returns
        -------
        (sha_of_target_branch, sha_after_update, sha_after_rewrite)
        """
        repo = self._repo
        source_branch = merge_request.source_branch
        target_branch = merge_request.target_branch
        assert source_repo_url != repo.remote_url
        if source_repo_url is None and source_branch == target_branch:
            raise CannotMerge("source and target branch seem to coincide!")

        branch_update_done = commits_rewrite_done = False
        try:
            initial_mr_sha = merge_request.sha
            updated_sha = self.fuse(
                source_branch, target_branch, source_repo_url=source_repo_url
            )
            branch_update_done = True
            # The fuse above fetches origin again, so we are now safe to fetch
            # the sha from the remote target branch.
            target_sha = repo.get_commit_hash("origin/" + target_branch)
            if updated_sha == target_sha:
                raise CannotMerge(
                    f"these changes already exist in branch `{target_branch}`"
                )
            final_sha = self.modify_commits(merge_request)
            final_sha = final_sha or updated_sha
            commits_rewrite_done = True
            branch_was_modified = final_sha != initial_mr_sha
            self.synchronize_mr_with_local_changes(
                merge_request, branch_was_modified, source_repo_url, skip_ci=skip_ci
            )
        except git.GitError as err:
            # A failure to clean up probably means something is fucked with the git repo
            # and likely explains any previous failure, so it will better to just
            # raise a GitError
            if source_branch != self.project.default_branch:
                repo.checkout_branch(self.project.default_branch)
                repo.remove_branch(source_branch)

            if not branch_update_done:
                raise CannotMerge(
                    "There are merge conflicts and the rebase failed. Please fix"
                    " them manually before assigning them back to Marge"
                ) from err
            if not commits_rewrite_done:
                raise CannotMerge(
                    "failed on filter-branch. Check bot logs for more details."
                ) from err
            raise
        return target_sha, updated_sha, final_sha

    def synchronize_mr_with_local_changes(
        self,
        merge_request: MergeRequest,
        branch_was_modified: bool,
        source_repo_url: Optional[str] = None,
        skip_ci: bool = False,
    ) -> None:
        if self._options.fusion is Fusion.gitlab_rebase:
            self.synchronize_using_gitlab_rebase(merge_request)
        else:
            self.push_force_to_mr(
                merge_request,
                branch_was_modified,
                source_repo_url=source_repo_url,
                skip_ci=skip_ci,
            )

    def push_force_to_mr(
        self,
        merge_request: MergeRequest,
        branch_was_modified: bool,
        source_repo_url: Optional[str] = None,
        skip_ci: bool = False,
    ) -> None:
        try:
            self._repo.push(
                merge_request.source_branch,
                source_repo_url=source_repo_url,
                force=True,
                skip_ci=skip_ci,
            )
        except git.GitError as err:
            if hasattr(err, "__cause__") and "(stale info)" in getattr(
                err.__cause__, "stderr", ""
            ):
                raise NeedsRebase("MR needs to be rebased.") from err

            def fetch_remote_branch() -> Branch:
                return Branch.fetch_by_name(
                    merge_request.source_project_id,
                    merge_request.source_branch,
                    self._api,
                )

            if branch_was_modified and fetch_remote_branch().protected:
                raise CannotMerge(
                    "The branch is protected and therefore cannot be rebased."
                    " Either relax the branch protection rules or opt for a different merge strategy."
                ) from err

            change_type = "merged" if self.opts.fusion == Fusion.merge else "rebased"
            raise CannotMerge(
                f"Failed to push {change_type} changes. Check bot logs for more details."
            ) from err

    def synchronize_using_gitlab_rebase(
        self, merge_request: MergeRequest, expected_sha: Optional[str] = None
    ) -> None:
        expected_sha = expected_sha or self._repo.get_commit_hash()
        try:
            merge_request.rebase()
        except MergeRequestRebaseFailed as err:
            raise CannotMerge(
                f"GitLab failed to rebase the branch saying: {err.args[0]}"
            ) from err
        except TimeoutError as err:
            raise CannotMerge("GitLab took too long to rebase the branch.") from err
        except gitlab.ApiError as err:
            branch = Branch.fetch_by_name(
                merge_request.source_project_id, merge_request.source_branch, self._api
            )
            if branch.protected:
                raise CannotMerge(
                    "The branch is protected and therefore cannot be rebased."
                    " Either relax the branch protection rules or opt for a different merge strategy."
                ) from err
            raise


def _get_reviewer_names_and_emails(
    commits: list[dict[str, Any]], approvals: Approvals, api: gitlab.Api
) -> list[str]:
    """Return a list ['A. Prover <a.prover@example.com', ...]` for `merge_request.`"""
    uids = approvals.approver_ids
    users = [User.fetch_by_id(uid, api) for uid in uids]
    self_reviewed = {commit["author_email"] for commit in commits} & {
        user.email for user in users
    }
    if self_reviewed and len(users) <= 1:
        raise CannotMerge("Commits require at least one independent reviewer.")
    return [f"{user.name} <{user.email}>" for user in users]


# pylint: disable=invalid-name
@enum.unique
class Fusion(enum.Enum):
    merge = 0
    rebase = 1
    gitlab_rebase = 2


@dataclasses.dataclass
class MergeJobOptions:
    add_tested: bool = False
    add_part_of: bool = False
    add_reviewers: bool = False
    keep_committers: bool = False
    keep_reviewers: bool = False
    reapprove: bool = False
    approval_timeout: Optional[datetime.timedelta] = dataclasses.field(
        default_factory=lambda: datetime.timedelta(seconds=0)
    )
    embargo: Optional[interval.IntervalUnion] = dataclasses.field(
        default_factory=IntervalUnion.empty
    )
    ci_timeout: Optional[datetime.timedelta] = dataclasses.field(
        default_factory=lambda: datetime.timedelta(minutes=15)
    )
    fusion: Fusion = Fusion.rebase
    use_no_ff_batches: bool = False
    use_merge_commit_batches: bool = False
    skip_ci_batches: bool = False
    skip_incomplete_pipelines: bool = False
    guarantee_final_pipeline: bool = False
    custom_allowed_approvers: Optional[list[str]] = None
    custom_required_approvals: int = 0
    hooks_directory: Optional[str] = None
    run_manual_jobs: bool = False
    sign_commits: bool = False

    def __post_init__(self) -> None:
        # Set any fields with None values to the result of its default_factory
        # callable.
        for field_ in dataclasses.fields(self):
            if getattr(self, field_.name) is None:
                if field_.default_factory != dataclasses.MISSING:
                    setattr(self, field_.name, field_.default_factory())

    @property
    def requests_commit_tagging(self) -> bool:
        return self.add_tested or self.add_part_of or self.add_reviewers


class CannotMerge(Exception):
    @property
    def reason(self) -> str:
        args = self.args
        if not args:
            return "Unknown reason!"

        return cast(str, args[0])


class SkipMerge(CannotMerge):
    pass


class NeedsRebase(CannotMerge):
    pass



from unittest.mock import Mock

import pytest

from marge.gitlab import GET, Api
from marge.project import AccessLevel, Project

INFO = {
    "id": 1234,
    "path_with_namespace": "cool/project",
    "ssh_url_to_repo": "ssh://blah.com/cool/project.git",
    "merge_requests_enabled": True,
    "default_branch": "master",
    "only_allow_merge_if_pipeline_succeeds": True,
    "only_allow_merge_if_all_discussions_are_resolved": False,
    "permissions": {
        "project_access": {"access_level": AccessLevel.developer.value},
        "group_access": {"access_level": AccessLevel.developer.value},
    },
}

GROUP_ACCESS = {
    "project_access": None,
    "group_access": {"access_level": AccessLevel.developer.value},
}

NONE_ACCESS = {"project_access": None, "group_access": None}


# pylint: disable=attribute-defined-outside-init,duplicate-code
class TestProject:
    def setup_method(self, _method):
        self.api = Mock(Api)

    def test_fetch_by_id(self):
        api = self.api
        api.call = Mock(return_value=INFO)

        project = Project.fetch_by_id(project_id=1234, api=api)

        api.call.assert_called_once_with(GET("/projects/1234"))
        assert project.info == INFO

    def test_properties(self):
        project = Project(api=self.api, info=INFO)
        assert project.id == 1234
        assert project.path_with_namespace == "cool/project"
        assert project.ssh_url_to_repo == "ssh://blah.com/cool/project.git"
        assert project.merge_requests_enabled is True
        assert project.only_allow_merge_if_pipeline_succeeds is True
        assert project.only_allow_merge_if_all_discussions_are_resolved is False
        assert project.access_level == AccessLevel.developer

    def test_group_access(self):
        project = Project(api=self.api, info=dict(INFO, permissions=GROUP_ACCESS))
        bad_project = Project(api=self.api, info=dict(INFO, permissions=NONE_ACCESS))
        assert project.access_level == AccessLevel.developer
        with pytest.raises(AssertionError):
            bad_project.access_level  # pylint: disable=pointless-statement



#!/usr/bin/env python3
"""Executable script to pass to git filter-branch --msgfilter to rewrite trailers.

This treats everything (stdin, stdout, env) at the level of raw bytes which are
assumed to be utf-8, or more specifically some ASCII superset, regardless of
(possibly broken) LOCALE settings.

"""

import collections
import os
import re
import sys
from typing import NoReturn

STDIN = sys.stdin.buffer
STDOUT = sys.stdout.buffer
STDERR = sys.stderr.buffer


def die(msg: bytes) -> NoReturn:
    STDERR.write(b"ERROR: ")
    STDERR.write(msg)
    sys.exit(1)


def drop_trailing_newlines(lines: list[bytes]) -> None:
    while lines and not lines[-1]:
        del lines[-1]


def remove_duplicates(trailers: list[bytes]) -> list[bytes]:
    return list(collections.OrderedDict((t, None) for t in trailers).keys())


def rework_commit_message(
    commit_message: bytes, trailers: list[bytes], keep_trailers: bool
) -> bytes:
    if not commit_message:
        die(b"Expected a non-empty commit message")

    trailer_names = [trailer.split(b":", 1)[0].lower() for trailer in trailers]

    filtered_lines = [
        line.rstrip()
        for line in commit_message.split(b"\n")
        if keep_trailers or line.split(b":", 1)[0].lower() not in trailer_names
    ]

    reworked_lines = filtered_lines[:]

    drop_trailing_newlines(reworked_lines)
    while len(reworked_lines) > 1 and re.match(rb"^[A-Z][\w-]+: ", reworked_lines[-1]):
        trailers.insert(0, reworked_lines.pop())
    if not reworked_lines:
        die(b"Your commit message seems to consist only of Trailers: " + commit_message)

    drop_trailing_newlines(reworked_lines)

    non_empty_trailers = remove_duplicates(
        [t for t in trailers if t.split(b": ", 1)[1].strip()]
    )
    if non_empty_trailers:
        reworked_lines += [b""] + non_empty_trailers
    reworked_lines += [b""]
    return b"\n".join(reworked_lines)


def main() -> int:
    trailers = os.environb[b"TRAILERS"].split(b"\n") if os.environb[b"TRAILERS"] else []
    keep_trailers: bool = os.environb.get(b"KEEP_TRAILERS", b"False") == b"True"
    assert all(b":" in trailer for trailer in trailers), trailers
    original_commit_message = STDIN.read().strip()
    new_commit_message = rework_commit_message(
        original_commit_message, trailers, keep_trailers
    )
    STDOUT.write(new_commit_message)
    return 0


if __name__ == "__main__":
    sys.exit(main())



# pylint: disable=protected-access
import datetime
from unittest.mock import ANY, create_autospec, patch

import pytest

import marge.git
import marge.gitlab
import marge.project
import marge.user
from marge.batch_job import BatchMergeJob, CannotBatch
from marge.gitlab import GET
from marge.job import CannotMerge, MergeJobOptions
from marge.merge_request import MergeRequest
from marge.pipeline import Pipeline
from tests.gitlab_api_mock import MockLab, Ok, commit


class TestBatchJob:
    @pytest.fixture(params=[True, False])
    def fork(self, request):
        return request.param

    @pytest.fixture()
    def mocklab(self, fork):
        return MockLab(fork=fork)

    @pytest.fixture()
    def api(self, mocklab):
        return mocklab.api

    def _mock_merge_request(self, **options):
        return create_autospec(
            marge.merge_request.MergeRequest, spec_set=True, **options
        )

    def _mock_pipeline(self, **options):
        return create_autospec(Pipeline, spec_set=True, **options)

    def get_batch_merge_job(self, api, mocklab, **batch_merge_kwargs):
        project_id = mocklab.project_info["id"]
        merge_request_iid = mocklab.merge_request_info["iid"]

        merge_request = MergeRequest.fetch_by_iid(project_id, merge_request_iid, api)

        params = {
            "api": api,
            "user": marge.user.User.myself(api),
            "project": marge.project.Project.fetch_by_id(project_id, api),
            "repo": create_autospec(
                marge.git.Repo(
                    remote_url="",
                    local_path="",
                    user_name="",
                    user_email="",
                    ssh_key_file=None,
                    timeout=datetime.timedelta(0),
                    reference="",
                    keep_committers=False,
                ),
                spec_set=True,
            ),
            "options": MergeJobOptions(),
            "merge_requests": [merge_request],
            "batch_branch_name": "test_branch_name",
        }
        params.update(batch_merge_kwargs)
        return BatchMergeJob(**params)

    def test_remove_local_batch_branch(self, api, mocklab):
        repo = create_autospec(marge.git.Repo, spec_set=True)
        batch_merge_job = self.get_batch_merge_job(api, mocklab, repo=repo)
        batch_merge_job.remove_local_batch_branch()

        repo.remove_branch.assert_called_once_with(batch_merge_job.batch_branch_name)

    def test_remove_remote_batch_branch(self, api, mocklab):
        with patch(
            "marge.batch_job.branch.Branch.delete_by_name"
        ) as delete_batch_branch:
            batch_merge_job = self.get_batch_merge_job(api, mocklab)
            batch_merge_job.remove_remote_batch_branch()

        delete_batch_branch.assert_called_once_with(
            project_id=batch_merge_job._project.id,
            branch=batch_merge_job.batch_branch_name,
            api=batch_merge_job._api,
        )

    def test_remove_remote_batch_branch_ignore_missing(self, api, mocklab):
        with patch(
            "marge.batch_job.branch.Branch.delete_by_name",
            side_effect=marge.gitlab.NotFound("missing branch"),
        ) as delete_batch_branch:
            batch_merge_job = self.get_batch_merge_job(api, mocklab)
            batch_merge_job.remove_remote_batch_branch()
            delete_batch_branch.assert_called_once_with(
                project_id=batch_merge_job._project.id,
                branch=batch_merge_job.batch_branch_name,
                api=batch_merge_job._api,
            )

    def test_close_batch_mr(self, api, mocklab):
        with (
            patch("marge.batch_job.MergeRequest") as mr_class,
            patch("marge.batch_job.Pipeline") as pipeline_class,
        ):
            batch_mr = self._mock_merge_request()
            mr_class.search.return_value = [batch_mr]

            batch_pipeline = self._mock_pipeline(status="running")
            pipeline_class.pipelines_by_merge_request.return_value = [batch_pipeline]

            batch_merge_job = self.get_batch_merge_job(api, mocklab)
            batch_merge_job.close_batch_mr()

            params = {
                "author_id": batch_merge_job._user.id,
                "labels": batch_merge_job.batch_branch_name,
                "state": "opened",
                "order_by": "created_at",
                "sort": "desc",
            }
            mr_class.search.assert_called_once_with(
                api=ANY, project_id=ANY, params=params
            )
            batch_mr.close.assert_called_once()
            pipeline_class.pipelines_by_merge_request.assert_called_once_with(
                ANY, ANY, ANY
            )
            batch_pipeline.cancel.assert_called_once()

    def test_create_batch_mr(self, api, mocklab):
        with patch("marge.batch_job.MergeRequest") as mr_class:
            batch_mr = self._mock_merge_request()
            mr_class.create.return_value = batch_mr

            batch_merge_job = self.get_batch_merge_job(api, mocklab)
            target_branch = "master"
            r_batch_mr = batch_merge_job.create_batch_mr(target_branch)

            params = {
                "source_branch": batch_merge_job.batch_branch_name,
                "target_branch": target_branch,
                "title": "Marge Bot Batch MR - DO NOT TOUCH",
                "labels": batch_merge_job.batch_branch_name,
            }
            mr_class.create.assert_called_once_with(
                api=ANY, project_id=ANY, params=params
            )
            assert r_batch_mr is batch_mr

    def test_get_mrs_with_common_target_branch(self, api, mocklab):
        master_mrs = [
            self._mock_merge_request(target_branch="master"),
            self._mock_merge_request(target_branch="master"),
        ]
        non_master_mrs = [
            self._mock_merge_request(target_branch="non_master"),
            self._mock_merge_request(target_branch="non_master"),
        ]
        batch_merge_job = self.get_batch_merge_job(
            api, mocklab, merge_requests=non_master_mrs + master_mrs
        )
        r_maser_mrs = batch_merge_job.get_mrs_with_common_target_branch("master")
        assert r_maser_mrs == master_mrs

    @patch.object(BatchMergeJob, "get_mr_ci_status")
    def test_ensure_mergeable_mr_ci_not_ok(self, bmj_get_mr_ci_status, api, mocklab):
        batch_merge_job = self.get_batch_merge_job(api, mocklab)
        bmj_get_mr_ci_status.return_value = (
            "failed",
            "See pipeline http://link/.",
            12345,
        )
        merge_request = self._mock_merge_request(
            assignee_ids=[batch_merge_job._user.id],
            state="opened",
            draft=False,
            squash_on_merge=False,
        )
        merge_request.fetch_approvals.return_value.sufficient = True
        with pytest.raises(CannotBatch) as exc_info:
            batch_merge_job.ensure_mergeable_mr(merge_request)

        assert (
            str(exc_info.value)
            == "This MR has not passed CI. See pipeline http://link/."
        )

    def test_push_batch(self, api, mocklab):
        batch_merge_job = self.get_batch_merge_job(api, mocklab)
        batch_merge_job.push_batch()
        batch_merge_job._repo.push.assert_called_once_with(
            batch_merge_job.batch_branch_name, force=True
        )

    def test_merge_batch(self, api, mocklab):
        batch_merge_job = self.get_batch_merge_job(api, mocklab)
        target_branch = "master"
        source_branch = mocklab.merge_request_info["source_branch"]
        batch_merge_job.merge_batch(target_branch, source_branch, no_ff=False)
        batch_merge_job._repo.fast_forward.assert_called_once_with(
            target_branch, source_branch
        )

    def test_merge_batch_with_no_ff_enabled(self, api, mocklab):
        batch_merge_job = self.get_batch_merge_job(api, mocklab)
        target_branch = "master"
        source_branch = mocklab.merge_request_info["source_branch"]
        batch_merge_job.merge_batch(target_branch, source_branch, no_ff=True)
        batch_merge_job._repo.merge.assert_called_once_with(
            target_branch, source_branch, "--no-ff"
        )
        batch_merge_job._repo.fast_forward.assert_not_called()

    def test_ensure_mr_not_changed(self, api, mocklab):
        with patch("marge.batch_job.MergeRequest") as mr_class:
            batch_merge_job = self.get_batch_merge_job(api, mocklab)
            merge_request = self._mock_merge_request()
            changed_merge_request = self._mock_merge_request()
            mr_class.fetch_by_iid.return_value = changed_merge_request

            with pytest.raises(CannotMerge):
                batch_merge_job.ensure_mr_not_changed(merge_request)

            mr_class.fetch_by_iid.assert_called_once_with(
                merge_request.project_id, merge_request.iid, batch_merge_job._api
            )

    def test_fuse_mr_when_target_branch_was_moved(self, api, mocklab):
        batch_merge_job = self.get_batch_merge_job(api, mocklab)
        merge_request = self._mock_merge_request(target_branch="master")
        with pytest.raises(CannotBatch) as exc_info:
            batch_merge_job.accept_mr(merge_request, "abc")
        assert str(exc_info.value) == "Someone was naughty and by-passed marge"

    def test_fuse_mr_when_source_branch_was_moved(self, api, mocklab):
        batch_merge_job = self.get_batch_merge_job(api, mocklab)
        merge_request = self._mock_merge_request(
            source_project_id=mocklab.merge_request_info["source_project_id"],
            target_branch="master",
            source_branch=mocklab.merge_request_info["source_branch"],
        )

        api.add_transition(
            GET(
                f'/projects/{mocklab.merge_request_info["source_project_id"]}/'
                f"repository/branches/useless_new_feature"
            ),
            Ok({"commit": commit(commit_id="abc", status="running")}),
        )

        with pytest.raises(CannotMerge) as exc_info:
            batch_merge_job.accept_mr(merge_request, mocklab.initial_master_sha)

        assert (
            str(exc_info.value)
            == "Someone pushed to branch while we were trying to merge"
        )



import abc
import dataclasses
import logging as log
from typing import TYPE_CHECKING, Any, Callable, Optional, Union, cast

import gitlab
import requests


class Api:
    def __init__(self, gitlab_url: str, auth_token: str) -> None:
        self._auth_token = auth_token
        self._api_base_url = gitlab_url.rstrip("/") + "/api/v4"
        self._gitlab = gitlab.Gitlab(gitlab_url, auth_token)

    def call(
        self, command: "Command", sudo: Optional[int] = None
    ) -> Union[bool, dict[str, Any], list[dict[str, Any]]]:
        method = command.method
        url = self._api_base_url + command.endpoint
        kwargs: dict[str, Any] = {}
        if sudo:
            kwargs["sudo"] = sudo
        log.debug("REQUEST: %s %s %s %r", method, url, sudo, command.call_args)
        # Timeout to prevent indefinitely hanging requests. 60s is very conservative,
        # but should be short enough to not cause any practical annoyances. We just
        # crash rather than retry since marge-bot should be run in a restart loop anyway.
        try:
            response = self._gitlab.http_request(
                method,
                command.endpoint,
                timeout=60,
                retry_transient_errors=True,
                **command.call_args,
                **kwargs,
            )
        except requests.exceptions.Timeout as err:
            log.error("Request timeout: %s", err)
            raise
        except gitlab.GitlabHttpError as err:
            code = err.response_code
            errors = {
                400: BadRequest,
                401: Unauthorized,
                403: Forbidden,
                404: NotFound,
                405: MethodNotAllowed,
                409: Conflict,
                422: Unprocessable,
                500: InternalServerError,
            }

            if code in errors:
                raise errors[code](*err.args) from err

            if code is not None and 500 < code < 600:
                raise InternalServerError(*err.args) from err

            raise UnexpectedError(*err.args) from err

        log.debug("RESPONSE CODE: %s", response.status_code)
        log.debug("RESPONSE BODY: %r", response.content)

        if response.status_code == 202:
            return True  # Accepted

        if response.status_code == 204:
            return True  # NoContent

        return command.extract(response.json()) if command.extract else response.json()

    def collect_all_pages(self, get_command: "GET") -> list[dict[str, Any]]:
        result: list[dict[str, Any]] = []
        fetch_again, page_no = True, 1
        while fetch_again:
            page = self.call(get_command.for_page(page_no))
            if page:
                if TYPE_CHECKING:
                    assert isinstance(page, list)
                result.extend(page)
                page_no += 1
            else:
                fetch_again = False

        return result


def from_singleton_list(
    fun: Optional[Callable[[dict[str, Any]], Any]] = None,
) -> Callable[[list[dict[str, Any]]], Any]:
    def extractor(response_list: list[dict[str, Any]]) -> Any:
        assert isinstance(response_list, list), type(response_list)
        assert len(response_list) <= 1, len(response_list)
        if not response_list:
            return None
        if fun is None:
            return response_list[0]
        return fun(response_list[0])

    return extractor


@dataclasses.dataclass(frozen=True)
class Command(abc.ABC):
    endpoint: str
    args: dict[str, Any] = dataclasses.field(default_factory=dict)
    extract: Optional[Callable[[list[dict[str, Any]]], dict[str, Any]]] = None

    @property
    def method(self) -> str:
        return self.__class__.__name__

    @property
    def call_args(self) -> dict[str, Any]:
        return {"post_data": self.args}


class GET(Command):
    @property
    def call_args(self) -> dict[str, Any]:
        return {"query_data": self.args}

    def for_page(self, page_no: int) -> "GET":
        args = self.args
        return dataclasses.replace(self, args=dict(args, page=page_no, per_page=100))


class PUT(Command):
    pass


class POST(Command):
    pass


class DELETE(Command):
    pass


class ApiError(gitlab.GitlabHttpError):
    @property
    def error_message(self) -> str:
        return self._error_message

    @error_message.setter
    def error_message(self, error: Union[str, dict[str, Any]]) -> None:
        # Might be as deep as {"message": {"base": ["real message"]}}
        if isinstance(error, dict) and "base" in error:
            error = cast(dict[str, Any], error.get("base"))

        if isinstance(error, list) and len(error) == 1:
            error = error[0]

        if TYPE_CHECKING:
            assert isinstance(error, str)
        self._error_message = error


class BadRequest(ApiError):
    pass


class Unauthorized(ApiError):
    pass


class Forbidden(ApiError):
    pass


class NotFound(ApiError):
    pass


class MethodNotAllowed(ApiError):
    pass


class Conflict(ApiError):
    pass


class Unprocessable(ApiError):
    pass


class InternalServerError(ApiError):
    pass


class UnexpectedError(ApiError):
    pass


class Resource(abc.ABC):
    def __init__(self, api: Api, info: dict[str, Any]):
        self._info = info
        self._api = api

    @property
    def info(self) -> dict[str, Any]:
        return self._info

    @property
    @abc.abstractmethod
    def id(self) -> Union[int, str]:  # pylint: disable=invalid-name
        ...

    @property
    def api(self) -> Api:
        return self._api

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._api}, {self.info})"



import dataclasses
import datetime
import os
import pathlib
import re
import shlex
import subprocess
from unittest import mock

import pytest

import marge.git
from marge import keepcommitters
from marge.git import GIT_SSH_COMMAND

GIT_CLONE_COMMAND = "git clone --origin=origin --filter=blob:none"


# pylint: disable=attribute-defined-outside-init
@mock.patch("marge.git._run")
class TestRepo:
    def setup_method(self, _method):
        self.repo = marge.git.Repo(
            remote_url="ssh://git@git.foo.com/some/repo.git",
            local_path="/tmp/local/path",
            user_name="bart simpson",
            user_email="bart.simpson@gmail.com",
            ssh_key_file=None,
            timeout=datetime.timedelta(seconds=1),
            reference=None,
            keep_committers=False,
        )

    def test_clone(self, mocked_run):
        self.repo.clone()
        assert get_calls(mocked_run) == [
            f"{GIT_CLONE_COMMAND} ssh://git@git.foo.com/some/repo.git /tmp/local/path"
        ]

    def test_configure(self, mocked_run):
        self.repo.configure()
        assert get_calls(mocked_run) == [
            "git -C /tmp/local/path config user.email bart.simpson@gmail.com",
            "git -C /tmp/local/path config user.name 'bart simpson'",
        ]

    def test_fetch(self, mocked_run):
        self.repo.fetch("source", "ssh://git@git.foo.com/forked/repo.git")
        assert get_calls(mocked_run) == [
            "git -C /tmp/local/path remote rm source",
            "git -C /tmp/local/path remote add source ssh://git@git.foo.com/forked/repo.git",
            "git -C /tmp/local/path config remote.source.partialclonefilter blob:none",
            "git -C /tmp/local/path fetch --prune source",
        ]

    def test_rebase_success(self, mocked_run):
        self.repo.rebase("feature_branch", "master_of_the_universe")

        assert get_calls(mocked_run) == [
            "git -C /tmp/local/path fetch --prune origin",
            "git -C /tmp/local/path checkout -B feature_branch origin/feature_branch --",
            "git -C /tmp/local/path rebase origin/master_of_the_universe",
            "git -C /tmp/local/path rev-parse HEAD",
        ]

    def test_merge_success(self, mocked_run):
        self.repo.merge("feature_branch", "master_of_the_universe")

        assert get_calls(mocked_run) == [
            "git -C /tmp/local/path fetch --prune origin",
            "git -C /tmp/local/path checkout -B feature_branch origin/feature_branch --",
            "git -C /tmp/local/path merge origin/master_of_the_universe",
            "git -C /tmp/local/path rev-parse HEAD",
        ]

    def test_reviewer_tagging_success(self, mocked_run):
        self.repo.tag_with_trailer(
            trailer_name="Reviewed-by",
            trailer_values=["John Simon <john@invalid>"],
            branch="feature_branch",
            start_commit="origin/master_of_the_universe",
            keep_trailers=False,
        )

        rewrite, parse = get_calls(mocked_run)
        pattern = "".join(
            [
                "git -C /tmp/local/path filter-branch --force ",
                "--msg-filter.*John Simon <john@invalid>.*origin/master_of_the_universe..feature_branch",
            ]
        )
        assert re.match(pattern, rewrite)
        assert parse == "git -C /tmp/local/path rev-parse HEAD"

    def test_reviewer_tagging_failure(self, mocked_run):
        def fail_on_filter_branch(*args, **unused_kwargs):
            if "filter-branch" in args:
                raise subprocess.CalledProcessError(returncode=1, cmd="git rebase blah")
            if "rev-parse" in args or "reset" in args:
                return mock.Mock()
            raise NotImplementedError("Unexpected call:", args)

        mocked_run.side_effect = fail_on_filter_branch

        try:
            self.repo.tag_with_trailer(
                trailer_name="Reviewed-by",
                branch="feature_branch",
                start_commit="origin/master_of_the_universe",
                trailer_values=["John Simon <john@invalid.com>"],
                keep_trailers=False,
            )
        except marge.git.GitError:
            pass
        else:
            assert False
        rewrite, check, abort = get_calls(mocked_run)
        assert "filter-branch" in rewrite
        assert check == "git -C /tmp/local/path rev-parse refs/original/refs/heads/"
        assert (
            abort
            == "git -C /tmp/local/path reset --hard refs/original/refs/heads/feature_branch"
        )

    def test_sign_success(self, mocked_run):
        self.repo.sign_commits(
            branch="feature_branch", start_commit="origin/master_of_the_universe"
        )

        rewrite, parse = get_calls(mocked_run)
        pattern = "".join(
            [
                "git -C /tmp/local/path filter-branch --force ",
                "--commit-filter 'git commit-tree -S \\$@' ",
                "--env-filter "
                '\'GIT_COMMITTER_EMAIL="bart.simpson@gmail.com" '
                'GIT_COMMITTER_NAME="bart simpson"\' ',
                "origin/master_of_the_universe..feature_branch",
            ]
        )
        assert re.match(pattern, rewrite)
        assert parse == "git -C /tmp/local/path rev-parse HEAD"

    def test_rebase_same_branch(self, mocked_run):
        with pytest.raises(AssertionError):
            self.repo.rebase("branch", "branch")

        assert get_calls(mocked_run) == []

    def test_merge_same_branch(self, mocked_run):
        with pytest.raises(AssertionError):
            self.repo.merge("branch", "branch")

        assert get_calls(mocked_run) == []

    def test_remove_branch(self, mocked_run):
        self.repo.remove_branch("some_branch", new_current_branch="devel")
        assert get_calls(mocked_run) == ["git -C /tmp/local/path branch -D some_branch"]

    def test_remove_branch_default(self, mocked_run):
        self.repo.remove_branch("some_branch")
        assert get_calls(mocked_run) == ["git -C /tmp/local/path branch -D some_branch"]

    def test_remove_master_branch_fails(self, unused_mocked_run):
        with pytest.raises(AssertionError):
            self.repo.remove_branch("meister", new_current_branch="meister")

    def test_push_force(self, mocked_run):
        mocked_run.return_value = mocked_stdout(b"")
        self.repo.push("my_branch", force=True)
        assert get_calls(mocked_run) == [
            "git -C /tmp/local/path checkout my_branch --",
            "git -C /tmp/local/path diff-index --quiet HEAD",
            "git -C /tmp/local/path ls-files --others",
            "git -C /tmp/local/path push --force-with-lease origin my_branch:my_branch",
        ]

    def test_push_force_fails_on_dirty(self, mocked_run):
        def fail_on_diff_index(*args, **unused_kwargs):
            if "diff-index" in args:
                raise subprocess.CalledProcessError(
                    returncode=1, cmd="git diff-index blah"
                )

        mocked_run.side_effect = fail_on_diff_index

        with pytest.raises(marge.git.GitError):
            self.repo.push("my_branch", force=True)

        assert get_calls(mocked_run) == [
            "git -C /tmp/local/path checkout my_branch --",
            "git -C /tmp/local/path diff-index --quiet HEAD",
        ]

    def test_push_force_fails_on_untracked(self, mocked_run):
        def fail_on_ls_files(*args, **unused_kwargs):
            if "ls-files" in args:
                return mocked_stdout("some_file.txt\nanother_file.py")
            return None

        mocked_run.side_effect = fail_on_ls_files

        with pytest.raises(marge.git.GitError):
            self.repo.push("my_branch", force=True)

        assert get_calls(mocked_run) == [
            "git -C /tmp/local/path checkout my_branch --",
            "git -C /tmp/local/path diff-index --quiet HEAD",
            "git -C /tmp/local/path ls-files --others",
        ]

    def test_get_commit_hash(self, mocked_run):
        mocked_run.return_value = mocked_stdout(b"deadbeef")

        commit_hash = self.repo.get_commit_hash()
        assert commit_hash == "deadbeef"

        assert get_calls(mocked_run) == ["git -C /tmp/local/path rev-parse HEAD"]
        self.repo.get_commit_hash(rev="master")
        assert get_calls(mocked_run)[-1] == "git -C /tmp/local/path rev-parse master"

    def test_passes_ssh_key(self, mocked_run):
        repo = dataclasses.replace(self.repo, ssh_key_file="/foo/id_rsa")
        repo.configure()
        git_ssh = (
            f"GIT_SSH_COMMAND='{GIT_SSH_COMMAND} -F /dev/null -o IdentitiesOnly=yes -i "
            f"/foo/id_rsa'"
        )
        assert get_calls(mocked_run) == [
            f"{git_ssh} git -C /tmp/local/path config user.email bart.simpson@gmail.com",
            f"{git_ssh} git -C /tmp/local/path config user.name 'bart simpson'",
            f"{git_ssh} git -C /tmp/local/path config gpg.format ssh",
            f"{git_ssh} git -C /tmp/local/path config user.signingKey /foo/id_rsa",
        ]

    def test_passes_reference_repo(self, mocked_run):
        repo = dataclasses.replace(self.repo, reference="/foo/reference_repo")
        repo.clone()
        assert get_calls(mocked_run) == [
            f"{GIT_CLONE_COMMAND} --reference=/foo/reference_repo ssh://git@git.foo.com/some/repo.git "
            + "/tmp/local/path"
        ]

    def test_keep_committers(self, mocked_run):
        repo = dataclasses.replace(self.repo, keep_committers=True)
        repo.rebase("feature_branch", "master_of_the_universe")
        assert get_calls(mocked_run) == [
            "git -C /tmp/local/path fetch --prune origin",
            "git -C /tmp/local/path checkout -B feature_branch origin/feature_branch --",
            f"git -c 'rebase.instructionFormat=%s%nexec python3 {keepcommitters.__file__} "
            + "%H' -C /tmp/local/path rebase origin/master_of_the_universe",
            "git -C /tmp/local/path rev-parse HEAD",
        ]


def get_calls(mocked_run):
    return [bashify(call) for call in mocked_run.call_args_list]


def bashify(call):
    args, kwargs = call
    args = [shlex.quote(arg) for arg in args]
    env = kwargs.get("env") or {}
    default_env = os.environ.copy()
    default_env["FILTER_BRANCH_SQUELCH_WARNING"] = "1"
    alt_env = [
        shlex.quote(k) + "=" + shlex.quote(v)
        for k, v in set(env.items()) - set(default_env.items())
    ]
    return " ".join(alt_env + args)


def mocked_stdout(stdout):
    return subprocess.CompletedProcess(["blah", "args"], 0, stdout, None)


def _filter_test(message, trailer_name, trailer_values, keep_trailers):
    script = marge.git._filter_branch_script(  # pylint: disable=protected-access
        trailer_name, trailer_values, keep_trailers
    )
    result = subprocess.check_output(
        [b"sh", b"-c", script.encode("utf-8")],
        input=message.encode("utf-8"),
        stderr=subprocess.STDOUT,
    )
    return result.decode("utf-8")


def test_filter():
    assert _filter_test("Some Stuff", "Tested-by", [], False) == "Some Stuff\n"
    assert _filter_test("Some Stuff\n", "Tested-by", [], False) == "Some Stuff\n"
    assert (
        _filter_test(
            "Some Stuff", "Tested-by", ["T. Estes <testes@example.com>"], False
        )
        == """Some Stuff

Tested-by: T. Estes <testes@example.com>
"""
    )

    test_commit_message = r"""Fix: bug in BLah.

Some stuff.
Some More stuff (really? Yeah: really!)

Reviewed-by: R. Viewer <rviewer@example.com>
Reviewed-by: R. Viewer <rviewer@example.com>
Signed-off-by: Stephen Offer <soffer@example.com>
"""
    with_tested_by = _filter_test(
        test_commit_message, "Tested-by", ["T. Estes <testes@example.com>"], False
    )
    assert with_tested_by == """Fix: bug in BLah.

Some stuff.
Some More stuff (really? Yeah: really!)

Reviewed-by: R. Viewer <rviewer@example.com>
Signed-off-by: Stephen Offer <soffer@example.com>
Tested-by: T. Estes <testes@example.com>
"""
    with_new_reviewed_by = _filter_test(
        with_tested_by,
        "Reviewed-by",
        ["Roger Ebert <ebert@example.com>", "John Simon <simon@example.com>"],
        False,
    )
    assert with_new_reviewed_by == """Fix: bug in BLah.

Some stuff.
Some More stuff (really? Yeah: really!)

Signed-off-by: Stephen Offer <soffer@example.com>
Tested-by: T. Estes <testes@example.com>
Reviewed-by: Roger Ebert <ebert@example.com>
Reviewed-by: John Simon <simon@example.com>
"""
    with_keep_reviewed_by = _filter_test(
        with_tested_by,
        "Reviewed-by",
        ["Roger Ebert <ebert@example.com>", "John Simon <simon@example.com>"],
        True,
    )
    assert with_keep_reviewed_by == """Fix: bug in BLah.

Some stuff.
Some More stuff (really? Yeah: really!)

Reviewed-by: R. Viewer <rviewer@example.com>
Signed-off-by: Stephen Offer <soffer@example.com>
Tested-by: T. Estes <testes@example.com>
Reviewed-by: Roger Ebert <ebert@example.com>
Reviewed-by: John Simon <simon@example.com>
"""
    assert (
        _filter_test("Test: frobnificator", "Tested-by", [], False)
        == "Test: frobnificator\n"
    )
    assert _filter_test(
        "Test: frobnificator", "Tested-by", ["T. Estes <testes@example.com>"], False
    ) == (
        """Test: frobnificator

Tested-by: T. Estes <testes@example.com>
"""
    )


def test_filter_fails_on_empty_commit_messages():
    with pytest.raises(subprocess.CalledProcessError) as exc_info:
        _filter_test("", "", [], False)
    assert exc_info.value.output == b"ERROR: Expected a non-empty commit message"


def test_filter_fails_on_commit_messages_that_are_empty_apart_from_trailers():
    with pytest.raises(subprocess.CalledProcessError) as exc_info:
        _filter_test(
            "Tested-by: T. Estes <testes@example.com>",
            "Tested-by",
            ["T. Estes <testes@example.com>"],
            False,
        )
    assert exc_info.value.output == b"".join(
        [
            b"ERROR: Your commit message seems to consist only of ",
            b"Trailers: Tested-by: T. Estes <testes@example.com>",
        ]
    )

    with pytest.raises(subprocess.CalledProcessError) as exc_info:
        _filter_test("", "Tested-by", ["T. Estes <testes@example.com>"], False)
    assert exc_info.value.output == b"ERROR: Expected a non-empty commit message"


def test_filter_ignore_first_line_trailer_in_commit_message_if_not_set():
    assert (
        _filter_test(
            "Tested-by: T. Estes <testes@example.com>",
            "Reviewed-by",
            ["John Simon <john@invalid>"],
            False,
        )
        == """Tested-by: T. Estes <testes@example.com>

Reviewed-by: John Simon <john@invalid>
"""
    )


def _commit_blob(repo, index):
    blob = pathlib.Path(repo.remote_url) / f"blob{index}"
    with blob.open("wb") as file:
        file.truncate(1024)
    repo.git("-C", repo.remote_url, "add", blob.name)
    repo.git("-C", repo.remote_url, "commit", "-m", f"Commit {index}", blob.name)


def test_blobless_rebase(_git_user_config, tmp_repo):
    """Test rebase against a "remote" repo and a local blobless checkout."""
    _commit_blob(tmp_repo, "1")
    _commit_blob(tmp_repo, "2a")
    tmp_repo.git("-C", tmp_repo.remote_url, "checkout", "-b", "feature", "HEAD~1")
    _commit_blob(tmp_repo, "2b")

    tmp_repo.clone()
    tmp_repo.rebase("feature", "main")



from unittest.mock import Mock

from marge.commit import Commit
from marge.gitlab import GET, Api

INFO = {
    "id": "6104942438c14ec7bd21c6cd5bd995272b3faff6",
    "short_id": "6104942438c",
    "title": "Sanitize for network graph",
    "author_name": "randx",
    "author_email": "dmitriy.zaporozhets@gmail.com",
    "committer_name": "Dmitriy",
    "committer_email": "dmitriy.zaporozhets@gmail.com",
    "created_at": "2012-09-20T09:06:12+03:00",
    "message": "Sanitize for network graph",
    "committed_date": "2012-09-20T09:06:12+03:00",
    "authored_date": "2012-09-20T09:06:12+03:00",
    "parent_ids": ["ae1d9fb46aa2b07ee9836d49862ec4e2c46fbbba"],
    "stats": {"additions": 15, "deletions": 10, "total": 25},
    "status": "running",
}


# pylint: disable=attribute-defined-outside-init
class TestProjectWithCommits:
    def setup_method(self, _method):
        self.api = Mock(Api)

    def test_fetch_by_id(self):
        api = self.api
        api.call = Mock(return_value=INFO)

        commit = Commit.fetch_by_id(project_id=1234, sha=INFO["id"], api=api)

        api.call.assert_called_once_with(
            GET(
                "/projects/1234/repository/commits/6104942438c14ec7bd21c6cd5bd995272b3faff6"
            )
        )
        assert commit.info == INFO

    def test_last_on_branch(self):
        self.api.call.side_effect = lambda *_, **__: {"commit": INFO}
        Commit.last_on_branch(project_id=1234, branch="foobar-branch", api=self.api)
        self.api.call.assert_called_once_with(
            GET("/projects/1234/repository/branches/foobar-branch")
        )

    def test_last_on_branch_encoding(self):
        self.api.call.side_effect = lambda *_, **__: {"commit": INFO}
        Commit.last_on_branch(project_id=1234, branch="foo/bar", api=self.api)
        self.api.call.assert_called_once_with(
            GET("/projects/1234/repository/branches/foo%2Fbar")
        )

    def test_properties(self):
        commit = Commit(api=self.api, info=INFO)
        assert commit.id == "6104942438c14ec7bd21c6cd5bd995272b3faff6"
        assert commit.short_id == "6104942438c"
        assert commit.title == "Sanitize for network graph"
        assert commit.author_name == "randx"
        assert commit.author_email == "dmitriy.zaporozhets@gmail.com"
        assert commit.status == "running"
